import glob
import os
import struct

import numpy as np

from yt.frontends.sph.io import IOHandlerSPH
from yt.frontends.tipsy.definitions import npart_mapping
from yt.utilities.lib.particle_kdtree_tools import generate_smoothing_length
from yt.utilities.logger import ytLogger as mylog


class IOHandlerTipsyBinary(IOHandlerSPH):
    _dataset_type = "tipsy"
    _vector_fields = {"Coordinates": 3, "Velocity": 3, "Velocities": 3}

    _pdtypes = None  # dtypes, to be filled in later
    _aux_pdtypes = None  # auxiliary files' dtypes

    _ptypes = ("Gas", "DarkMatter", "Stars")

    _aux_fields = None
    _fields = (
        ("Gas", "Mass"),
        ("Gas", "Coordinates"),
        ("Gas", "Velocities"),
        ("Gas", "Density"),
        ("Gas", "Temperature"),
        ("Gas", "Epsilon"),
        ("Gas", "Metals"),
        ("Gas", "Phi"),
        ("DarkMatter", "Mass"),
        ("DarkMatter", "Coordinates"),
        ("DarkMatter", "Velocities"),
        ("DarkMatter", "Epsilon"),
        ("DarkMatter", "Phi"),
        ("Stars", "Mass"),
        ("Stars", "Coordinates"),
        ("Stars", "Velocities"),
        ("Stars", "Metals"),
        ("Stars", "FormationTime"),
        ("Stars", "Epsilon"),
        ("Stars", "Phi"),
    )

    def __init__(self, *args, **kwargs):
        self._aux_fields = []
        super().__init__(*args, **kwargs)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _fill_fields(self, fields, vals, hsml, mask, data_file):
        if mask is None:
            size = 0
        elif isinstance(mask, slice):
            size = vals[fields[0]].size
        else:
            size = mask.sum()
        rv = {}
        for field in fields:
            mylog.debug("Allocating %s values for %s", size, field)
            if field in self._vector_fields:
                rv[field] = np.empty((size, 3), dtype="float64")
                if size == 0:
                    continue
                rv[field][:, 0] = vals[field]["x"][mask]
                rv[field][:, 1] = vals[field]["y"][mask]
                rv[field][:, 2] = vals[field]["z"][mask]
            elif field == "smoothing_length":
                rv[field] = hsml[mask]
            else:
                rv[field] = np.empty(size, dtype="float64")
                if size == 0:
                    continue
                rv[field][:] = vals[field][mask]
            if field == "Coordinates":
                eps = np.finfo(rv[field].dtype).eps
                for i in range(3):
                    rv[field][:, i] = np.clip(
                        rv[field][:, i],
                        self.ds.domain_left_edge[i].v + eps,
                        self.ds.domain_right_edge[i].v - eps,
                    )
        return rv

    def _read_particle_coords(self, chunks, ptf):
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        chunksize = self.ds.index.chunksize
        for data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            poff = data_file.field_offsets
            tp = data_file.total_particles
            f = open(data_file.filename, "rb")
            for ptype in sorted(ptf, key=lambda a, poff=poff: poff.get(a, -1)):
                if data_file.total_particles[ptype] == 0:
                    continue
                f.seek(poff[ptype])
                total = 0
                while total < tp[ptype]:
                    count = min(chunksize, tp[ptype] - total)
                    p = np.fromfile(f, self._pdtypes[ptype], count=count)
                    total += p.size
                    d = [p["Coordinates"][ax].astype("float64") for ax in "xyz"]
                    del p
                    if ptype == self.ds._sph_ptypes[0]:
                        hsml = self._read_smoothing_length(data_file, count)
                    else:
                        hsml = 0.0
                    yield ptype, d, hsml

    @property
    def hsml_filename(self):
        return f"{self.ds.parameter_filename}-{'hsml'}"

    def _generate_smoothing_length(self, index):
        if os.path.exists(self.hsml_filename):
            with open(self.hsml_filename, "rb") as f:
                file_hash = struct.unpack("q", f.read(struct.calcsize("q")))[0]
            if file_hash != self.ds._file_hash:
                os.remove(self.hsml_filename)
            else:
                return
        positions = []
        for data_file in index.data_files:
            for _, ppos in self._yield_coordinates(
                data_file, needed_ptype=self.ds._sph_ptypes[0]
            ):
                positions.append(ppos)
        if positions == []:
            return
        kdtree = index.kdtree
        positions = np.concatenate(positions)[kdtree.idx]
        hsml = generate_smoothing_length(positions, kdtree, self.ds._num_neighbors)
        hsml = hsml[np.argsort(kdtree.idx)]
        dtype = self._pdtypes["Gas"]["Coordinates"][0]
        with open(self.hsml_filename, "wb") as f:
            f.write(struct.pack("q", self.ds._file_hash))
            f.write(hsml.astype(dtype).tobytes())

    def _read_smoothing_length(self, data_file, count):
        dtype = self._pdtypes["Gas"]["Coordinates"][0]
        with open(self.hsml_filename, "rb") as f:
            f.seek(struct.calcsize("q") + data_file.start * dtype.itemsize)
            hsmls = np.fromfile(f, dtype, count=count)
        return hsmls.astype("float64")

    def _get_smoothing_length(self, data_file, dtype, shape):
        return self._read_smoothing_length(data_file, shape[0])

    def _read_particle_data_file(self, data_file, ptf, selector=None):
        from numpy.lib.recfunctions import append_fields

        return_data = {}

        poff = data_file.field_offsets
        aux_fields_offsets = self._calculate_particle_offsets_aux(data_file)
        tp = data_file.total_particles
        f = open(data_file.filename, "rb")

        # we need to open all aux files for chunking to work
        aux_fh = {}
        for afield in self._aux_fields:
            aux_fh[afield] = open(data_file.filename + "." + afield, "rb")

        for ptype, field_list in sorted(ptf.items(), key=lambda a: poff.get(a[0], -1)):
            if data_file.total_particles[ptype] == 0:
                continue
            f.seek(poff[ptype])
            afields = list(set(field_list).intersection(self._aux_fields))
            count = min(self.ds.index.chunksize, tp[ptype])
            p = np.fromfile(f, self._pdtypes[ptype], count=count)
            auxdata = []
            for afield in afields:
                aux_fh[afield].seek(aux_fields_offsets[afield][ptype])
                if isinstance(self._aux_pdtypes[afield], np.dtype):
                    auxdata.append(
                        np.fromfile(
                            aux_fh[afield], self._aux_pdtypes[afield], count=count
                        )
                    )
                else:
                    par = self.ds.parameters
                    nlines = 1 + par["nsph"] + par["ndark"] + par["nstar"]
                    aux_fh[afield].seek(0)
                    sh = aux_fields_offsets[afield][ptype]
                    sf = nlines - count - sh
                    if tp[ptype] > 0:
                        aux = np.genfromtxt(
                            aux_fh[afield], skip_header=sh, skip_footer=sf
                        )
                        if aux.ndim < 1:
                            aux = np.array([aux])
                        auxdata.append(aux)
            if afields:
                p = append_fields(p, afields, auxdata)
            if ptype == "Gas":
                hsml = self._read_smoothing_length(data_file, count)
            else:
                hsml = 0.0
            if selector is None or getattr(selector, "is_all_data", False):
                mask = slice(None, None, None)
            else:
                x = p["Coordinates"]["x"].astype("float64")
                y = p["Coordinates"]["y"].astype("float64")
                z = p["Coordinates"]["z"].astype("float64")
                mask = selector.select_points(x, y, z, hsml)
                del x, y, z
            if mask is None:
                continue
            tf = self._fill_fields(field_list, p, hsml, mask, data_file)
            for field in field_list:
                return_data[(ptype, field)] = tf.pop(field)

        # close all file handles
        f.close()
        for fh in list(aux_fh.values()):
            fh.close()

        return return_data

    def _update_domain(self, data_file):
        """
        This method is used to determine the size needed for a box that will
        bound the particles.  It simply finds the largest position of the
        whole set of particles, and sets the domain to +/- that value.
        """
        ds = data_file.ds
        ind = 0
        # NOTE:
        #  We hardcode this value here because otherwise we get into a
        #  situation where we require the existence of index before we
        #  can successfully instantiate it, or where we are calling it
        #  from within its instantiation.
        #
        #  Because this value is not propagated later on, and does not
        #  impact the construction of the bitmap indices, it should be
        #  acceptable to just use a reasonable number here.
        chunksize = 64**3
        # Check to make sure that the domain hasn't already been set
        # by the parameter file
        if np.all(np.isfinite(ds.domain_left_edge)) and np.all(
            np.isfinite(ds.domain_right_edge)
        ):
            return
        with open(data_file.filename, "rb") as f:
            ds.domain_left_edge = 0
            ds.domain_right_edge = 0
            f.seek(ds._header_offset)
            mi = np.array([1e30, 1e30, 1e30], dtype="float64")
            ma = -np.array([1e30, 1e30, 1e30], dtype="float64")
            for ptype in self._ptypes:
                # We'll just add the individual types separately
                count = data_file.total_particles[ptype]
                if count == 0:
                    continue
                stop = ind + count
                while ind < stop:
                    c = min(chunksize, stop - ind)
                    pp = np.fromfile(f, dtype=self._pdtypes[ptype], count=c)
                    np.minimum(
                        mi,
                        [
                            pp["Coordinates"]["x"].min(),
                            pp["Coordinates"]["y"].min(),
                            pp["Coordinates"]["z"].min(),
                        ],
                        mi,
                    )
                    np.maximum(
                        ma,
                        [
                            pp["Coordinates"]["x"].max(),
                            pp["Coordinates"]["y"].max(),
                            pp["Coordinates"]["z"].max(),
                        ],
                        ma,
                    )
                    ind += c
        # We extend by 1%.
        DW = ma - mi
        mi -= 0.01 * DW
        ma += 0.01 * DW
        ds.domain_left_edge = ds.arr(mi, "code_length")
        ds.domain_right_edge = ds.arr(ma, "code_length")
        ds.domain_width = DW = ds.domain_right_edge - ds.domain_left_edge
        ds.unit_registry.add(
            "unitary", float(DW.max() * DW.units.base_value), DW.units.dimensions
        )

    def _yield_coordinates(self, data_file, needed_ptype=None):
        with open(data_file.filename, "rb") as f:
            poff = data_file.field_offsets
            for ptype in self._ptypes:
                if ptype not in poff:
                    continue
                f.seek(poff[ptype])
                if needed_ptype is not None and ptype != needed_ptype:
                    continue
                # We'll just add the individual types separately
                count = data_file.total_particles[ptype]
                if count == 0:
                    continue
                pp = np.fromfile(f, dtype=self._pdtypes[ptype], count=count)
                mis = np.empty(3, dtype="float64")
                mas = np.empty(3, dtype="float64")
                for axi, ax in enumerate("xyz"):
                    mi = pp["Coordinates"][ax].min()
                    ma = pp["Coordinates"][ax].max()
                    mylog.debug("Spanning: %0.3e .. %0.3e in %s", mi, ma, ax)
                    mis[axi] = mi
                    mas[axi] = ma
                pos = np.empty((pp.size, 3), dtype="float64")
                for i, ax in enumerate("xyz"):
                    pos[:, i] = pp["Coordinates"][ax]
                yield ptype, pos

    def _count_particles(self, data_file):
        pcount = np.array(
            [
                data_file.ds.parameters["nsph"],
                data_file.ds.parameters["nstar"],
                data_file.ds.parameters["ndark"],
            ]
        )
        si, ei = data_file.start, data_file.end
        if None not in (si, ei):
            np.clip(pcount - si, 0, ei - si, out=pcount)
        ptypes = ["Gas", "Stars", "DarkMatter"]
        npart = dict(zip(ptypes, pcount))
        return npart

    @classmethod
    def _compute_dtypes(cls, field_dtypes, endian="<"):
        pds = {}
        for ptype, field in cls._fields:
            dtbase = field_dtypes.get(field, "f")
            ff = f"{endian}{dtbase}"
            if field in cls._vector_fields:
                dt = (field, [("x", ff), ("y", ff), ("z", ff)])
            else:
                dt = (field, ff)
            pds.setdefault(ptype, []).append(dt)
        pdtypes = {}
        for ptype in pds:
            pdtypes[ptype] = np.dtype(pds[ptype])
        return pdtypes

    def _create_dtypes(self, data_file):
        # We can just look at the particle counts.
        self._header_offset = data_file.ds._header_offset
        self._pdtypes = self._compute_dtypes(
            data_file.ds._field_dtypes, data_file.ds.endian
        )
        self._field_list = []
        for ptype, field in self._fields:
            if data_file.total_particles[ptype] == 0:
                # We do not want out _pdtypes to have empty particles.
                self._pdtypes.pop(ptype, None)
                continue
            self._field_list.append((ptype, field))

        if "Gas" in self._pdtypes.keys():
            self._field_list.append(("Gas", "smoothing_length"))

        # Find out which auxiliaries we have and what is their format
        tot_parts = np.sum(
            [
                data_file.ds.parameters["nsph"],
                data_file.ds.parameters["nstar"],
                data_file.ds.parameters["ndark"],
            ]
        )
        endian = data_file.ds.endian
        self._aux_pdtypes = {}
        self._aux_fields = []
        for f in glob.glob(data_file.filename + ".*"):
            afield = f.rsplit(".")[-1]
            filename = data_file.filename + "." + afield
            if not os.path.exists(filename):
                continue
            if afield in ["log", "parameter", "kdtree"]:
                # Amiga halo finder makes files like this we need to ignore
                continue
            self._aux_fields.append(afield)
        skip_afields = []
        for afield in self._aux_fields:
            filename = data_file.filename + "." + afield
            # We need to do some fairly ugly detection to see what format the
            # auxiliary files are in.  They can be either ascii or binary, and
            # the binary files can be either floats, ints, or doubles.  We're
            # going to use a try-catch cascade to determine the format.
            filesize = os.stat(filename).st_size
            dtype = np.dtype(endian + "i4")
            tot_parts_from_file = np.fromfile(filename, dtype, count=1)
            if tot_parts_from_file != tot_parts:
                with open(filename, "rb") as f:
                    header_nparts = f.readline()
                    try:
                        header_nparts = int(header_nparts)
                    except ValueError:
                        skip_afields.append(afield)
                        continue
                    if int(header_nparts) != tot_parts:
                        raise RuntimeError
                self._aux_pdtypes[afield] = "ascii"
            elif (filesize - 4) / 8 == tot_parts:
                self._aux_pdtypes[afield] = np.dtype([("aux", endian + "d")])
            elif (filesize - 4) / 4 == tot_parts:
                if afield.startswith("i"):
                    self._aux_pdtypes[afield] = np.dtype([("aux", endian + "i")])
                else:
                    self._aux_pdtypes[afield] = np.dtype([("aux", endian + "f")])
            else:
                skip_afields.append(afield)
        for afield in skip_afields:
            self._aux_fields.remove(afield)
        # Add the auxiliary fields to each ptype we have
        for ptype in self._ptypes:
            if any(ptype == field[0] for field in self._field_list):
                self._field_list += [(ptype, afield) for afield in self._aux_fields]
        return self._field_list

    def _identify_fields(self, data_file):
        return self._field_list, {}

    def _calculate_particle_offsets(self, data_file, pcounts):
        # This computes the offsets for each particle type into a "data_file."
        # Note that the term "data_file" here is a bit overloaded, and also refers to a
        # "chunk" of particles inside a data file.
        # data_file.start represents the *particle count* that we should start at.
        #
        # At this point, data_file will have the total number of particles
        # that this chunk represents located in the property total_particles.
        # Because in tipsy files the particles are stored sequentially, we can
        # figure out where each one starts.
        # We first figure out the global offsets, then offset them by the count
        # and size of each individual particle type.
        field_offsets = {}
        # Initialize pos to the point the first particle type would start
        pos = data_file.ds._header_offset
        global_offsets = {}
        field_offsets = {}
        for ptype in self._ptypes:
            if ptype not in self._pdtypes:
                # This means we don't have any, I think, and so we shouldn't
                # stick it in the offsets.
                continue
            # Note that much of this will be computed redundantly; future
            # refactorings could fix this.
            global_offsets[ptype] = pos
            size = self._pdtypes[ptype].itemsize
            npart = self.ds.parameters[npart_mapping[ptype]]
            # Get the offset into just this particle type, and start at data_file.start
            if npart > data_file.start:
                field_offsets[ptype] = pos + size * data_file.start
            pos += npart * size
        return field_offsets

    def _calculate_particle_offsets_aux(self, data_file):
        aux_fields_offsets = {}
        params = self.ds.parameters
        for afield in self._aux_fields:
            aux_fields_offsets[afield] = {}
            if isinstance(self._aux_pdtypes[afield], np.dtype):
                pos = 4  # i4
                size = np.dtype(self._aux_pdtypes[afield]).itemsize
            else:
                pos = 1
                size = 1
            for i, ptype in enumerate(self._ptypes):
                if data_file.total_particles[ptype] == 0:
                    continue
                elif params[npart_mapping[ptype]] > self.ds.index.chunksize:
                    for j in range(i):
                        npart = params[npart_mapping[self._ptypes[j]]]
                        if npart > self.ds.index.chunksize:
                            pos += npart * size
                    pos += data_file.start * size
                aux_fields_offsets[afield][ptype] = pos
                pos += data_file.total_particles[ptype] * size
        return aux_fields_offsets



import os
import weakref
from collections import defaultdict
from functools import cached_property, partial

import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
)
from yt.data_objects.static_output import ParticleDataset
from yt.frontends.gadget.data_structures import _fix_unit_ordering
from yt.frontends.gadget_fof.fields import GadgetFOFFieldInfo, GadgetFOFHaloFieldInfo
from yt.frontends.halo_catalog.data_structures import HaloCatalogFile, HaloDataset
from yt.funcs import only_on_root, setdefaultattr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities.cosmology import Cosmology
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py


class GadgetFOFParticleIndex(ParticleIndex):
    def _calculate_particle_count(self):
        """
        Calculate the total number of each type of particle.
        """
        self.particle_count = {
            ptype: sum(d.total_particles[ptype] for d in self.data_files)
            for ptype in self.ds.particle_types_raw
        }

    def _calculate_particle_index_starts(self):
        # Halo indices are not saved in the file, so we must count by hand.
        # File 0 has halos 0 to N_0 - 1, file 1 has halos N_0 to N_0 + N_1 - 1, etc.
        particle_count = defaultdict(int)
        offset_count = 0
        for data_file in self.data_files:
            data_file.index_start = {
                ptype: particle_count[ptype] for ptype in data_file.total_particles
            }
            data_file.offset_start = offset_count
            for ptype in data_file.total_particles:
                particle_count[ptype] += data_file.total_particles[ptype]
            offset_count += data_file.total_offset

        self._halo_index_start = {
            ptype: np.array(
                [data_file.index_start[ptype] for data_file in self.data_files]
            )
            for ptype in self.ds.particle_types_raw
        }

    def _calculate_file_offset_map(self):
        # After the FOF is performed, a load-balancing step redistributes halos
        # and then writes more fields.  Here, for each file, we create a list of
        # files which contain the rest of the redistributed particles.
        ifof = np.array(
            [data_file.total_particles["Group"] for data_file in self.data_files]
        )
        isub = np.array([data_file.total_offset for data_file in self.data_files])
        subend = isub.cumsum()
        fofend = ifof.cumsum()
        istart = np.digitize(fofend - ifof, subend - isub) - 1
        iend = np.clip(np.digitize(fofend, subend), 0, ifof.size - 2)
        for i, data_file in enumerate(self.data_files):
            data_file.offset_files = self.data_files[istart[i] : iend[i] + 1]

    def _detect_output_fields(self):
        field_list = []
        units = {}
        found_fields = {
            ptype: False for ptype, pnum in self.particle_count.items() if pnum > 0
        }

        for data_file in self.data_files:
            fl, _units = self.io._identify_fields(data_file)
            units.update(_units)
            field_list.extend([f for f in fl if f not in field_list])
            for ptype in found_fields:
                found_fields[ptype] |= data_file.total_particles[ptype]
            if all(found_fields.values()):
                break

        self.field_list = field_list
        ds = self.dataset
        ds.particle_types = tuple({pt for pt, ds in field_list})
        ds.field_units.update(units)
        ds.particle_types_raw = ds.particle_types

    def _setup_filenames(self):
        template = self.ds.filename_template
        ndoms = self.ds.file_count
        cls = self.ds._file_class
        self.data_files = [
            cls(self.ds, self.io, template % {"num": i}, i, frange=None)
            for i in range(ndoms)
        ]

    def _setup_data_io(self):
        super()._setup_data_io()
        self._calculate_particle_count()
        self._calculate_particle_index_starts()
        self._calculate_file_offset_map()


class GadgetFOFHDF5File(HaloCatalogFile):
    def __init__(self, ds, io, filename, file_id, frange):
        with h5py.File(filename, mode="r") as f:
            self.header = {str(field): val for field, val in f["Header"].attrs.items()}
            self.group_length_sum = (
                f["Group/GroupLen"][()].sum() if "Group/GroupLen" in f else 0
            )
            self.group_subs_sum = (
                f["Group/GroupNsubs"][()].sum() if "Group/GroupNsubs" in f else 0
            )
        self.total_ids = self.header["Nids_ThisFile"]
        self.total_offset = 0
        super().__init__(ds, io, filename, file_id, frange)

    def _read_particle_positions(self, ptype, f=None):
        """
        Read all particle positions in this file.
        """

        if f is None:
            close = True
            f = h5py.File(self.filename, mode="r")
        else:
            close = False

        pos = f[ptype][f"{ptype}Pos"][()].astype("float64")

        if close:
            f.close()

        return pos


class GadgetFOFDataset(ParticleDataset):
    _index_class = GadgetFOFParticleIndex
    _file_class = GadgetFOFHDF5File
    _field_info_class = GadgetFOFFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="gadget_fof_hdf5",
        index_order=None,
        index_filename=None,
        unit_base=None,
        units_override=None,
        unit_system="cgs",
    ):
        if unit_base is not None and "UnitLength_in_cm" in unit_base:
            # We assume this is comoving, because in the absence of comoving
            # integration the redshift will be zero.
            unit_base["cmcm"] = 1.0 / unit_base["UnitLength_in_cm"]
        self._unit_base = unit_base
        if units_override is not None:
            raise RuntimeError(
                "units_override is not supported for GadgetFOFDataset. "
                + "Use unit_base instead."
            )
        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            index_order=index_order,
            index_filename=index_filename,
            unit_system=unit_system,
        )

    def add_field(self, *args, **kwargs):
        super().add_field(*args, **kwargs)
        self._halos_ds.add_field(*args, **kwargs)

    @property
    def halos_field_list(self):
        return self._halos_ds.field_list

    @property
    def halos_derived_field_list(self):
        return self._halos_ds.derived_field_list

    @cached_property
    def _halos_ds(self):
        return GadgetFOFHaloDataset(self)

    def _setup_classes(self):
        super()._setup_classes()
        self.halo = partial(GadgetFOFHaloContainer, ds=self._halos_ds)

    def _parse_parameter_file(self):
        with h5py.File(self.parameter_filename, mode="r") as f:
            self.parameters = {
                str(field): val for field, val in f["Header"].attrs.items()
            }

        self.dimensionality = 3
        self.refine_by = 2

        # Set standard values
        self.domain_left_edge = np.zeros(3, "float64")
        self.domain_right_edge = np.ones(3, "float64") * self.parameters["BoxSize"]
        self.domain_dimensions = np.ones(3, "int32")
        self.cosmological_simulation = 1
        self._periodicity = (True, True, True)
        self.current_redshift = self.parameters["Redshift"]
        self.omega_lambda = self.parameters["OmegaLambda"]
        self.omega_matter = self.parameters["Omega0"]
        self.hubble_constant = self.parameters["HubbleParam"]
        cosmology = Cosmology(
            hubble_constant=self.hubble_constant,
            omega_matter=self.omega_matter,
            omega_lambda=self.omega_lambda,
        )
        self.current_time = cosmology.t_from_z(self.current_redshift)

        prefix = os.path.abspath(
            os.path.join(
                os.path.dirname(self.parameter_filename),
                os.path.basename(self.parameter_filename).split(".", 1)[0],
            )
        )
        suffix = self.parameter_filename.rsplit(".", 1)[-1]
        self.filename_template = f"{prefix}.%(num)i.{suffix}"
        self.file_count = self.parameters["NumFiles"]
        self.particle_types = ("Group", "Subhalo")
        self.particle_types_raw = ("Group", "Subhalo")

    def _set_code_unit_attributes(self):
        # Set a sane default for cosmological simulations.
        if self._unit_base is None and self.cosmological_simulation == 1:
            only_on_root(mylog.info, "Assuming length units are in Mpc/h (comoving)")
            self._unit_base = {"length": (1.0, "Mpccm/h")}
        # The other same defaults we will use from the standard Gadget
        # defaults.
        unit_base = self._unit_base or {}

        if "length" in unit_base:
            length_unit = unit_base["length"]
        elif "UnitLength_in_cm" in unit_base:
            if self.cosmological_simulation == 0:
                length_unit = (unit_base["UnitLength_in_cm"], "cm")
            else:
                length_unit = (unit_base["UnitLength_in_cm"], "cmcm/h")
        else:
            raise RuntimeError
        length_unit = _fix_unit_ordering(length_unit)
        setdefaultattr(self, "length_unit", self.quan(length_unit[0], length_unit[1]))

        if "velocity" in unit_base:
            velocity_unit = unit_base["velocity"]
        elif "UnitVelocity_in_cm_per_s" in unit_base:
            velocity_unit = (unit_base["UnitVelocity_in_cm_per_s"], "cm/s")
        else:
            if self.cosmological_simulation == 0:
                velocity_unit = (1e5, "cm/s")
            else:
                velocity_unit = (1e5, "cm/s * sqrt(a)")
        velocity_unit = _fix_unit_ordering(velocity_unit)
        setdefaultattr(
            self, "velocity_unit", self.quan(velocity_unit[0], velocity_unit[1])
        )

        # We set hubble_constant = 1.0 for non-cosmology, so this is safe.
        # Default to 1e10 Msun/h if mass is not specified.
        if "mass" in unit_base:
            mass_unit = unit_base["mass"]
        elif "UnitMass_in_g" in unit_base:
            if self.cosmological_simulation == 0:
                mass_unit = (unit_base["UnitMass_in_g"], "g")
            else:
                mass_unit = (unit_base["UnitMass_in_g"], "g/h")
        else:
            # Sane default
            mass_unit = (1.0, "1e10*Msun/h")
        mass_unit = _fix_unit_ordering(mass_unit)
        setdefaultattr(self, "mass_unit", self.quan(mass_unit[0], mass_unit[1]))

        if "time" in unit_base:
            time_unit = unit_base["time"]
        elif "UnitTime_in_s" in unit_base:
            time_unit = (unit_base["UnitTime_in_s"], "s")
        else:
            tu = (self.length_unit / self.velocity_unit).to("yr/h")
            time_unit = (tu.d, tu.units)
        setdefaultattr(self, "time_unit", self.quan(time_unit[0], time_unit[1]))

    def __str__(self):
        return self.basename.split(".", 1)[0]

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        need_groups = ["Group", "Header", "Subhalo"]
        veto_groups = ["FOF"]
        valid = True
        try:
            fh = h5py.File(filename, mode="r")
            valid = all(ng in fh["/"] for ng in need_groups) and not any(
                vg in fh["/"] for vg in veto_groups
            )
            fh.close()
        except Exception:
            valid = False
            pass
        return valid


class GadgetFOFHaloParticleIndex(GadgetFOFParticleIndex):
    def __init__(self, ds, dataset_type):
        self.real_ds = weakref.proxy(ds.real_ds)
        super().__init__(ds, dataset_type)

    def _create_halo_id_table(self):
        """
        Create a list of halo start ids so we know which file
        contains particles for a given halo.  Note, the halo ids
        are distributed over all files and so the ids for a given
        halo are likely stored in a different file than the halo
        itself.
        """

        self._halo_id_number = np.array(
            [data_file.total_ids for data_file in self.data_files]
        )
        self._halo_id_end = self._halo_id_number.cumsum()
        self._halo_id_start = self._halo_id_end - self._halo_id_number

        self._group_length_sum = np.array(
            [data_file.group_length_sum for data_file in self.data_files]
        )

    def _detect_output_fields(self):
        field_list = []
        scalar_field_list = []
        units = {}
        found_fields = {
            ptype: False for ptype, pnum in self.particle_count.items() if pnum > 0
        }
        has_ids = False

        for data_file in self.data_files:
            fl, sl, idl, _units = self.io._identify_fields(data_file)
            units.update(_units)
            field_list.extend([f for f in fl if f not in field_list])
            scalar_field_list.extend([f for f in sl if f not in scalar_field_list])
            for ptype in found_fields:
                found_fields[ptype] |= data_file.total_particles[ptype]
            has_ids |= len(idl) > 0
            if all(found_fields.values()) and has_ids:
                break

        self.field_list = field_list
        self.scalar_field_list = scalar_field_list
        ds = self.dataset
        ds.scalar_field_list = scalar_field_list
        ds.particle_types = tuple({pt for pt, ds in field_list})
        ds.field_units.update(units)
        ds.particle_types_raw = ds.particle_types

    def _identify_base_chunk(self, dobj):
        pass

    def _read_particle_fields(self, fields, dobj, chunk=None):
        if len(fields) == 0:
            return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        fields_to_return = self.io._read_particle_selection(dobj, fields_to_read)
        return fields_to_return, fields_to_generate

    def _get_halo_file_indices(self, ptype, identifiers):
        return np.digitize(identifiers, self._halo_index_start[ptype], right=False) - 1

    def _get_halo_scalar_index(self, ptype, identifier):
        i_scalar = self._get_halo_file_indices(ptype, [identifier])[0]
        scalar_index = identifier - self._halo_index_start[ptype][i_scalar]
        return scalar_index

    def _get_halo_values(self, ptype, identifiers, fields, f=None):
        """
        Get field values for halos.  IDs are likely to be
        sequential (or at least monotonic), but not necessarily
        all within the same file.

        This does not do much to minimize file i/o, but with
        halos randomly distributed across files, there's not
        much more we can do.
        """

        # if a file is already open, don't open it again
        filename = None if f is None else f.filename

        data = defaultdict(lambda: np.empty(identifiers.size))
        i_scalars = self._get_halo_file_indices(ptype, identifiers)
        for i_scalar in np.unique(i_scalars):
            target = i_scalars == i_scalar
            scalar_indices = identifiers - self._halo_index_start[ptype][i_scalar]

            # only open file if it's not already open
            my_f = (
                f
                if self.data_files[i_scalar].filename == filename
                else h5py.File(self.data_files[i_scalar].filename, mode="r")
            )

            for field in fields:
                data[field][target] = my_f[os.path.join(ptype, field)][()][
                    scalar_indices[target]
                ]

            if self.data_files[i_scalar].filename != filename:
                my_f.close()

        return data

    def _setup_data_io(self):
        super()._setup_data_io()
        self._create_halo_id_table()


class GadgetFOFHaloDataset(HaloDataset):
    _index_class = GadgetFOFHaloParticleIndex
    _file_class = GadgetFOFHDF5File
    _field_info_class = GadgetFOFHaloFieldInfo

    def __init__(self, ds, dataset_type="gadget_fof_halo_hdf5"):
        super().__init__(ds, dataset_type)

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        # This class is not meant to be instantiated by yt.load()
        return False


class GadgetFOFHaloContainer(YTSelectionContainer):
    """
    Create a data container to get member particles and individual
    values from halos and subhalos. Halo mass, position, and
    velocity are set as attributes. Halo IDs are accessible
    through the field, "member_ids".  Other fields that are one
    value per halo are accessible as normal.  The field list for
    halo objects can be seen in `ds.halos_field_list`.

    Parameters
    ----------
    ptype : string
        The type of halo, either "Group" for the main halo or
        "Subhalo" for subhalos.
    particle_identifier : int or tuple of ints
        The halo or subhalo id.  If requesting a subhalo, the id
        can also be given as a tuple of the main halo id and
        subgroup id, such as (1, 4) for subgroup 4 of halo 1.

    Attributes
    ----------
    particle_identifier : int
        The id of the halo or subhalo.
    group_identifier : int
        For subhalos, the id of the enclosing halo.
    subgroup_identifier : int
        For subhalos, the relative id of the subhalo within
        the enclosing halo.
    particle_number : int
        Number of particles in the halo.
    mass : float
        Halo mass.
    position : array of floats
        Halo position.
    velocity : array of floats
        Halo velocity.

    Note
    ----
    Relevant Fields:

     * particle_number - number of particles
     * subhalo_number - number of subhalos
     * group_identifier - id of parent group for subhalos

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("gadget_halos/data/groups_298/fof_subhalo_tab_298.0.hdf5")

    >>> halo = ds.halo("Group", 0)
    >>> print(halo.mass)
    13256.5517578 code_mass
    >>> print(halo.position)
    [ 16.18603706   6.95965052  12.52694607] code_length
    >>> print(halo.velocity)
    [ 6943694.22793569  -762788.90647454  -794749.63819757] cm/s
    >>> print(halo["Group_R_Crit200"])
    [ 0.79668683] code_length

    >>> # particle ids for this halo
    >>> print(halo["member_ids"])
    [  723631.   690744.   854212. ...,   608589.   905551.  1147449.] dimensionless

    >>> # get the first subhalo of this halo
    >>> subhalo = ds.halo("Subhalo", (0, 0))
    >>> print(subhalo["member_ids"])
    [  723631.   690744.   854212. ...,   808362.   956359.  1248821.] dimensionless

    """

    _type_name = "halo"
    _con_args = ("ptype", "particle_identifier")
    _spatial = False
    # Do not register it to prevent .halo from being attached to all datasets
    _skip_add = True

    def __init__(self, ptype, particle_identifier, ds=None):
        if ptype not in ds.particle_types_raw:
            raise RuntimeError(
                f'Possible halo types are {ds.particle_types_raw}, supplied "{ptype}".'
            )

        self.ptype = ptype
        self._current_particle_type = ptype
        super().__init__(ds, {})

        if ptype == "Subhalo" and isinstance(particle_identifier, tuple):
            self.group_identifier, self.subgroup_identifier = particle_identifier
            my_data = self.index._get_halo_values(
                "Group", np.array([self.group_identifier]), ["GroupFirstSub"]
            )
            self.particle_identifier = np.int64(
                my_data["GroupFirstSub"][0] + self.subgroup_identifier
            )
        else:
            self.particle_identifier = particle_identifier

        if self.particle_identifier >= self.index.particle_count[ptype]:
            raise RuntimeError(
                "%s %d requested, but only %d %s objects exist."
                % (ptype, particle_identifier, self.index.particle_count[ptype], ptype)
            )

        # Find the file that has the scalar values for this halo.
        i_scalar = self.index._get_halo_file_indices(ptype, [self.particle_identifier])[
            0
        ]
        self.scalar_data_file = self.index.data_files[i_scalar]

        # index within halo arrays that corresponds to this halo
        self.scalar_index = self.index._get_halo_scalar_index(
            ptype, self.particle_identifier
        )

        halo_fields = [f"{ptype}Len"]
        if ptype == "Subhalo":
            halo_fields.append("SubhaloGrNr")
        my_data = self.index._get_halo_values(
            ptype, np.array([self.particle_identifier]), halo_fields
        )
        self.particle_number = np.int64(my_data[f"{ptype}Len"][0])

        if ptype == "Group":
            self.group_identifier = self.particle_identifier
            id_offset = 0
            # index of file that has scalar values for the group
            g_scalar = i_scalar
            group_index = self.scalar_index

        # If a subhalo, find the index of the parent.
        elif ptype == "Subhalo":
            self.group_identifier = np.int64(my_data["SubhaloGrNr"][0])

            # Find the file that has the scalar values for the parent group.
            g_scalar = self.index._get_halo_file_indices(
                "Group", [self.group_identifier]
            )[0]

            # index within halo arrays that corresponds to the paent group
            group_index = self.index._get_halo_scalar_index(
                "Group", self.group_identifier
            )

            my_data = self.index._get_halo_values(
                "Group",
                np.array([self.group_identifier]),
                ["GroupNsubs", "GroupFirstSub"],
            )
            self.subgroup_identifier = self.particle_identifier - np.int64(
                my_data["GroupFirstSub"][0]
            )
            parent_subhalos = my_data["GroupNsubs"][0]

            mylog.debug(
                "Subhalo %d is subgroup %s of %d in group %d.",
                self.particle_identifier,
                self.subgroup_identifier,
                parent_subhalos,
                self.group_identifier,
            )

            # ids of the sibling subhalos that come before this one
            if self.subgroup_identifier > 0:
                sub_ids = np.arange(
                    self.particle_identifier - self.subgroup_identifier,
                    self.particle_identifier,
                )
                my_data = self.index._get_halo_values(
                    "Subhalo", sub_ids, ["SubhaloLen"]
                )
                id_offset = my_data["SubhaloLen"].sum(dtype=np.int64)
            else:
                id_offset = 0

        # Calculate the starting index for the member particles.
        # First, add up all the particles in the earlier files.
        all_id_start = self.index._group_length_sum[:g_scalar].sum(dtype=np.int64)

        # Now add the halos in this file that come before.
        with h5py.File(self.index.data_files[g_scalar].filename, mode="r") as f:
            all_id_start += f["Group"]["GroupLen"][:group_index].sum(dtype=np.int64)

        # Add the subhalo offset.
        all_id_start += id_offset

        # indices of first and last files containing member particles
        i_start = (
            np.digitize([all_id_start], self.index._halo_id_start, right=False)[0] - 1
        )
        i_end = np.digitize(
            [all_id_start + self.particle_number], self.index._halo_id_end, right=True
        )[0]
        self.field_data_files = self.index.data_files[i_start : i_end + 1]

        # starting and ending indices for each file containing particles
        self.field_data_start = (
            all_id_start - self.index._halo_id_start[i_start : i_end + 1]
        ).clip(min=0)
        self.field_data_start = self.field_data_start.astype(np.int64)
        self.field_data_end = (
            all_id_start
            + self.particle_number
            - self.index._halo_id_start[i_start : i_end + 1]
        ).clip(max=self.index._halo_id_number[i_start : i_end + 1])
        self.field_data_end = self.field_data_end.astype(np.int64)

        for attr in ["mass", "position", "velocity"]:
            setattr(self, attr, self[self.ptype, f"particle_{attr}"][0])

    def __repr__(self):
        return "%s_%s_%09d" % (self.ds, self.ptype, self.particle_identifier)



contents = [
    (
        "Getting Started",
        [
            ("welcome/index.html", "Welcome to yt!", "What's yt all about?"),
            (
                "orientation/index.html",
                "yt Orientation",
                "Quickly get up and running with yt: zero to sixty.",
            ),
            (
                "help/index.html",
                "How to Ask for Help",
                "Some guidelines on how and where to ask for help with yt",
            ),
            (
                "workshop.html",
                "Workshop Tutorials",
                "Videos, slides and scripts from the 2012 workshop covering many "
                + "aspects of yt, from beginning to advanced.",
            ),
        ],
    ),
    (
        "Everyday yt",
        [
            (
                "analyzing/index.html",
                "Analyzing Data",
                "An overview of different ways to handle and process data: loading "
                + "data, using and manipulating objects and fields, examining and "
                + "manipulating particles, derived fields, generating processed data, "
                + "time series analysis.",
            ),
            (
                "visualizing/index.html",
                "Visualizing Data",
                "An overview of different ways to visualize data: making projections, "
                + "slices, phase plots, streamlines, and volume rendering; modifying "
                + "plots; the fixed resolution buffer.",
            ),
            (
                "interacting/index.html",
                "Interacting with yt",
                "Different ways -- scripting, GUIs, prompts, explorers -- to explore "
                + "your data.",
            ),
        ],
    ),
    (
        "Advanced Usage",
        [
            (
                "advanced/index.html",
                "Advanced yt usage",
                "Advanced topics: parallelism, debugging, ways to drive yt, "
                + "developing",
            ),
            (
                "getting_involved/index.html",
                "Getting Involved",
                "How to participate in the community, contribute code and share "
                + "scripts",
            ),
        ],
    ),
    (
        "Reference Materials",
        [
            (
                "cookbook/index.html",
                "The Cookbook",
                "A bunch of illustrated examples of how to do things",
            ),
            (
                "reference/index.html",
                "Reference Materials",
                "A list of all bundled fields, API documentation, the Change Log...",
            ),
            ("faq/index.html", "FAQ", "Frequently Asked Questions: answered for you!"),
        ],
    ),
]

heading_template = r"""
<h2>%s</h2>
<table class="contentstable" align="center">
%s
</table>
"""

subheading_template = r"""
  <tr valign="top">
    <td width="25%%">
      <p>
        <a href="%s">%s</a>
      </p>
    </td>
    <td width="75%%">
      <p class="linkdescr">%s</p>
    </td>
  </tr>
"""

t = ""
for heading, items in contents:
    s = ""
    for subheading in items:
        s += subheading_template % subheading
    t += heading_template % (heading, s)
print(t)



import sys

import numpy as np

from yt._typing import FieldType
from yt.fields.derived_field import ValidateParameter
from yt.fields.field_info_container import FieldInfoContainer
from yt.geometry.api import Geometry
from yt.units import dimensions

from .field_plugin_registry import register_field_plugin

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never

cgs_normalizations = {"gaussian": 4.0 * np.pi, "lorentz_heaviside": 1.0}


def get_magnetic_normalization(key: str) -> float:
    if key not in cgs_normalizations:
        raise ValueError(
            "Unknown magnetic normalization convention. "
            f"Got {key!r}, expected one of {tuple(cgs_normalizations)}"
        )
    return cgs_normalizations[key]


@register_field_plugin
def setup_magnetic_field_fields(
    registry: FieldInfoContainer, ftype: FieldType = "gas", slice_info=None
) -> None:
    ds = registry.ds

    unit_system = ds.unit_system
    pc = registry.ds.units.physical_constants

    axis_names = registry.ds.coordinates.axis_order

    if (ftype, f"magnetic_field_{axis_names[0]}") not in registry:
        return

    u = registry[ftype, f"magnetic_field_{axis_names[0]}"].units

    def mag_factors(dims):
        if dims == dimensions.magnetic_field_cgs:
            return getattr(ds, "_magnetic_factor", 4.0 * np.pi)
        elif dims == dimensions.magnetic_field_mks:
            return ds.units.physical_constants.mu_0

    def _magnetic_field_strength(field, data):
        xm = f"relative_magnetic_field_{axis_names[0]}"
        ym = f"relative_magnetic_field_{axis_names[1]}"
        zm = f"relative_magnetic_field_{axis_names[2]}"

        B2 = (data[ftype, xm]) ** 2 + (data[ftype, ym]) ** 2 + (data[ftype, zm]) ** 2

        return np.sqrt(B2)

    registry.add_field(
        (ftype, "magnetic_field_strength"),
        sampling_type="local",
        function=_magnetic_field_strength,
        validators=[ValidateParameter("bulk_magnetic_field")],
        units=u,
    )

    def _magnetic_energy_density(field, data):
        B = data[ftype, "magnetic_field_strength"]
        return 0.5 * B * B / mag_factors(B.units.dimensions)

    registry.add_field(
        (ftype, "magnetic_energy_density"),
        sampling_type="local",
        function=_magnetic_energy_density,
        units=unit_system["pressure"],
    )

    def _plasma_beta(field, data):
        return data[ftype, "pressure"] / data[ftype, "magnetic_energy_density"]

    registry.add_field(
        (ftype, "plasma_beta"), sampling_type="local", function=_plasma_beta, units=""
    )

    def _magnetic_pressure(field, data):
        return data[ftype, "magnetic_energy_density"]

    registry.add_field(
        (ftype, "magnetic_pressure"),
        sampling_type="local",
        function=_magnetic_pressure,
        units=unit_system["pressure"],
    )

    _magnetic_field_poloidal_magnitude = None
    _magnetic_field_toroidal_magnitude = None

    geometry: Geometry = registry.ds.geometry

    if geometry is Geometry.CARTESIAN:

        def _magnetic_field_poloidal_magnitude(field, data):
            B2 = (
                data[ftype, "relative_magnetic_field_x"]
                * data[ftype, "relative_magnetic_field_x"]
                + data[ftype, "relative_magnetic_field_y"]
                * data[ftype, "relative_magnetic_field_y"]
                + data[ftype, "relative_magnetic_field_z"]
                * data[ftype, "relative_magnetic_field_z"]
            )
            Bt2 = (
                data[ftype, "magnetic_field_spherical_phi"]
                * data[ftype, "magnetic_field_spherical_phi"]
            )
            return np.sqrt(B2 - Bt2)

    elif geometry is Geometry.CYLINDRICAL or geometry is Geometry.POLAR:

        def _magnetic_field_poloidal_magnitude(field, data):
            bm = data.get_field_parameter("bulk_magnetic_field")
            rax = axis_names.index("r")
            zax = axis_names.index("z")

            return np.sqrt(
                (data[ftype, "magnetic_field_r"] - bm[rax]) ** 2
                + (data[ftype, "magnetic_field_z"] - bm[zax]) ** 2
            )

        def _magnetic_field_toroidal_magnitude(field, data):
            ax = axis_names.find("theta")
            bm = data.get_field_parameter("bulk_magnetic_field")
            return data[ftype, "magnetic_field_theta"] - bm[ax]

    elif geometry is Geometry.SPHERICAL:

        def _magnetic_field_poloidal_magnitude(field, data):
            bm = data.get_field_parameter("bulk_magnetic_field")
            rax = axis_names.index("r")
            tax = axis_names.index("theta")

            return np.sqrt(
                (data[ftype, "magnetic_field_r"] - bm[rax]) ** 2
                + (data[ftype, "magnetic_field_theta"] - bm[tax]) ** 2
            )

        def _magnetic_field_toroidal_magnitude(field, data):
            ax = axis_names.find("phi")
            bm = data.get_field_parameter("bulk_magnetic_field")
            return data[ftype, "magnetic_field_phi"] - bm[ax]

    elif geometry is Geometry.GEOGRAPHIC or geometry is Geometry.INTERNAL_GEOGRAPHIC:
        # not implemented
        pass
    elif geometry is Geometry.SPECTRAL_CUBE:
        # nothing to be done
        pass
    else:
        assert_never(geometry)

    if _magnetic_field_poloidal_magnitude is not None:
        registry.add_field(
            (ftype, "magnetic_field_poloidal_magnitude"),
            sampling_type="local",
            function=_magnetic_field_poloidal_magnitude,
            units=u,
            validators=[
                ValidateParameter("normal"),
                ValidateParameter("bulk_magnetic_field"),
            ],
        )

    if _magnetic_field_toroidal_magnitude is not None:
        registry.add_field(
            (ftype, "magnetic_field_toroidal_magnitude"),
            sampling_type="local",
            function=_magnetic_field_toroidal_magnitude,
            units=u,
            validators=[
                ValidateParameter("normal"),
                ValidateParameter("bulk_magnetic_field"),
            ],
        )

    if geometry is Geometry.CARTESIAN:
        registry.alias(
            (ftype, "magnetic_field_toroidal_magnitude"),
            (ftype, "magnetic_field_spherical_phi"),
            units=u,
        )
        registry.alias(
            (ftype, "magnetic_field_toroidal"),
            (ftype, "magnetic_field_spherical_phi"),
            units=u,
            deprecate=("4.1.0", None),
        )
        registry.alias(
            (ftype, "magnetic_field_poloidal"),
            (ftype, "magnetic_field_spherical_theta"),
            units=u,
            deprecate=("4.1.0", None),
        )
    elif (
        geometry is Geometry.CYLINDRICAL
        or geometry is Geometry.POLAR
        or geometry is Geometry.SPHERICAL
    ):
        # These cases should be covered already, just check that they are
        assert (ftype, "magnetic_field_toroidal_magnitude") in registry
        assert (ftype, "magnetic_field_poloidal_magnitude") in registry
    elif geometry is Geometry.GEOGRAPHIC or geometry is Geometry.INTERNAL_GEOGRAPHIC:
        # not implemented
        pass
    elif geometry is Geometry.SPECTRAL_CUBE:
        # nothing to be done
        pass
    else:
        assert_never(Geometry)

    def _alfven_speed(field, data):
        B = data[ftype, "magnetic_field_strength"]
        return B / np.sqrt(mag_factors(B.units.dimensions) * data[ftype, "density"])

    registry.add_field(
        (ftype, "alfven_speed"),
        sampling_type="local",
        function=_alfven_speed,
        units=unit_system["velocity"],
    )

    def _mach_alfven(field, data):
        return data[ftype, "velocity_magnitude"] / data[ftype, "alfven_speed"]

    registry.add_field(
        (ftype, "mach_alfven"),
        sampling_type="local",
        function=_mach_alfven,
        units="dimensionless",
    )

    b_units = registry.ds.quan(1.0, u).units
    if dimensions.current_mks in b_units.dimensions.free_symbols:
        rm_scale = pc.qp.to("C", "SI") ** 3 / (4.0 * np.pi * pc.eps_0)
    else:
        rm_scale = pc.qp**3 / pc.clight
    rm_scale *= registry.ds.quan(1.0, "rad") / (
        2.0 * np.pi * pc.me**2 * pc.clight**3
    )
    rm_units = registry.ds.quan(1.0, "rad/m**2").units / unit_system["length"]

    def _rotation_measure(field, data):
        return (
            rm_scale
            * data[ftype, "magnetic_field_los"]
            * data[ftype, "El_number_density"]
        )

    registry.add_field(
        (ftype, "rotation_measure"),
        sampling_type="local",
        function=_rotation_measure,
        units=rm_units,
        validators=[ValidateParameter("axis", {"axis": [0, 1, 2]})],
    )


def setup_magnetic_field_aliases(registry, ds_ftype, ds_fields, ftype="gas"):
    r"""
    This routine sets up special aliases between dataset-specific magnetic
    fields and the default magnetic fields in yt so that unit conversions
    between different unit systems can be handled properly. This is only called
    from the `setup_fluid_fields` method (for grid dataset) or the
    `setup_gas_particle_fields` method (for particle dataset) of a frontend's
    :class:`FieldInfoContainer` instance.

    Parameters
    ----------
    registry : :class:`FieldInfoContainer`
        The field registry that these definitions will be installed into.
    ds_ftype : string
        The field type for the fields we're going to alias, e.g. "flash",
        "enzo", "athena", "PartType0", etc.
    ds_fields : list of strings or string
        The fields that will be aliased. For grid dataset, this should be a
        list of strings corresponding to the components of magnetic field. For
        particle dataset, this should be a single string corresponding to the
        vector magnetic field.
    ftype : string, optional
        The resulting field type of the fields. Default "gas".

    Examples
    --------
    >>> from yt.fields.magnetic_field import setup_magnetic_field_aliases
    >>> class PlutoFieldInfo(ChomboFieldInfo):
    ...     def setup_fluid_fields(self):
    ...         setup_magnetic_field_aliases(
    ...             self, "chombo", ["bx%s" % ax for ax in [1, 2, 3]]
    ...         )
    >>> class GizmoFieldInfo(GadgetFieldInfo):
    ...     def setup_gas_particle_fields(self):
    ...         setup_magnetic_field_aliases(
    ...             self, "PartType0", "MagneticField", ftype="PartType0"
    ...         )
    """
    unit_system = registry.ds.unit_system
    if isinstance(ds_fields, list):
        # If ds_fields is a list, we assume a grid dataset
        sampling_type = "local"
        ds_fields = [(ds_ftype, fd) for fd in ds_fields]
        ds_field = ds_fields[0]
    else:
        # Otherwise, we assume a particle dataset
        sampling_type = "particle"
        ds_field = (ds_ftype, ds_fields)
    if ds_field not in registry:
        return

    # Figure out the unit conversion to use
    if unit_system.base_units[dimensions.current_mks] is not None:
        to_units = unit_system["magnetic_field_mks"]
    else:
        to_units = unit_system["magnetic_field_cgs"]
    units = unit_system[to_units.dimensions]

    # Add fields
    if sampling_type in ["cell", "local"]:
        # Grid dataset case
        def mag_field_from_field(fd):
            def _mag_field(field, data):
                return data[fd].to(field.units)

            return _mag_field

        for ax, fd in zip(registry.ds.coordinates.axis_order, ds_fields):
            registry.add_field(
                (ftype, f"magnetic_field_{ax}"),
                sampling_type=sampling_type,
                function=mag_field_from_field(fd),
                units=units,
            )
    else:
        # Particle dataset case
        def mag_field_from_ax(ax):
            def _mag_field(field, data):
                return data[ds_field][:, "xyz".index(ax)]

            return _mag_field

        for ax in registry.ds.coordinates.axis_order:
            fname = f"particle_magnetic_field_{ax}"
            registry.add_field(
                (ds_ftype, fname),
                sampling_type=sampling_type,
                function=mag_field_from_ax(ax),
                units=units,
            )
            sph_ptypes = getattr(registry.ds, "_sph_ptypes", ())
            if ds_ftype in sph_ptypes:
                registry.alias((ftype, f"magnetic_field_{ax}"), (ds_ftype, fname))



# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
"""
Geometry selection routines.




"""


import numpy as np

cimport cython
cimport numpy as np
from libc.math cimport sqrt
from libc.stdlib cimport free, malloc

from yt.utilities.lib.bitarray cimport ba_get_value
from yt.utilities.lib.fnv_hash cimport c_fnv_hash as fnv_hash
from yt.utilities.lib.fp_utils cimport fclip, fmax, fmin, iclip
from yt.utilities.lib.grid_traversal cimport walk_volume
from yt.utilities.lib.volume_container cimport VolumeContainer

from .oct_container cimport Oct, OctreeContainer
from .oct_visitors cimport cind


cdef extern from "math.h":
    double exp(double x) nogil
    float expf(float x) nogil
    long double expl(long double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil
    double fmod(double x, double y) nogil
    double log2(double x) nogil
    long int lrint(double x) nogil
    double fabs(double x) nogil

# use this as an epsilon test for grids aligned with selector
# define here to avoid the gil later
cdef np.float64_t grid_eps = np.finfo(np.float64).eps
grid_eps = 0.0

cdef inline np.float64_t dot(np.float64_t* v1,
                             np.float64_t* v2) nogil:
    return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]

cdef inline np.float64_t norm(np.float64_t* v) nogil:
    return sqrt(dot(v, v))

# These routines are separated into a couple different categories:
#
#   * Routines for identifying intersections of an object with a bounding box
#   * Routines for identifying cells/points inside a bounding box that
#     intersect with an object
#   * Routines that speed up some type of geometric calculation

# First, bounding box / object intersection routines.
# These all respect the interface "dobj" and a set of left_edges, right_edges,
# sometimes also accepting level and mask information.

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def convert_mask_to_indices(np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
            int count, int transpose = 0):
    cdef int i, j, k, cpos
    cdef np.ndarray[np.int64_t, ndim=2] indices
    indices = np.zeros((count, 3), dtype='int64')
    cpos = 0
    for i in range(mask.shape[0]):
        for j in range(mask.shape[1]):
            for k in range(mask.shape[2]):
                if mask[i, j, k] == 1:
                    if transpose == 1:
                        indices[cpos, 0] = k
                        indices[cpos, 1] = j
                        indices[cpos, 2] = i
                    else:
                        indices[cpos, 0] = i
                        indices[cpos, 1] = j
                        indices[cpos, 2] = k
                    cpos += 1
    return indices


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef _mask_fill(np.ndarray[np.float64_t, ndim=1] out,
                np.int64_t offset,
                np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
                np.ndarray[cython.floating, ndim=3] vals):
    cdef np.int64_t count = 0
    cdef int i, j, k
    for i in range(mask.shape[0]):
        for j in range(mask.shape[1]):
            for k in range(mask.shape[2]):
                if mask[i, j, k] == 1:
                    out[offset + count] = vals[i, j, k]
                    count += 1
    return count

def mask_fill(np.ndarray[np.float64_t, ndim=1] out,
              np.int64_t offset,
              np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
              np.ndarray vals):
    if vals.dtype == np.float32:
        return _mask_fill[np.float32_t](out, offset, mask, vals)
    elif vals.dtype == np.float64:
        return _mask_fill[np.float64_t](out, offset, mask, vals)
    else:
        raise RuntimeError

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def points_in_cells(
        np.float64_t[:] cx,
        np.float64_t[:] cy,
        np.float64_t[:] cz,
        np.float64_t[:] dx,
        np.float64_t[:] dy,
        np.float64_t[:] dz,
        np.float64_t[:] px,
        np.float64_t[:] py,
        np.float64_t[:] pz):
    # Take a list of cells and particles and calculate which particles
    # are enclosed within one of the cells.  This is used for querying
    # particle fields on clump/contour objects.
    # We use brute force since the cells are a relatively unordered collection.

    cdef int p, c, n_p, n_c
    cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask

    n_p = px.size
    n_c = cx.size
    mask = np.zeros(n_p, dtype="bool")

    for p in range(n_p):
        for c in range(n_c):
            if (fabs(px[p] - cx[c]) <= 0.5 * dx[c] and
                fabs(py[p] - cy[c]) <= 0.5 * dy[c] and
                fabs(pz[p] - cz[c]) <= 0.5 * dz[c]):
                mask[p] = True
                break

    return mask

include "_selection_routines/selector_object.pxi"
include "_selection_routines/point_selector.pxi"
include "_selection_routines/sphere_selector.pxi"
include "_selection_routines/region_selector.pxi"
include "_selection_routines/cut_region_selector.pxi"
include "_selection_routines/disk_selector.pxi"
include "_selection_routines/cutting_plane_selector.pxi"
include "_selection_routines/slice_selector.pxi"
include "_selection_routines/ortho_ray_selector.pxi"
include "_selection_routines/ray_selector.pxi"
include "_selection_routines/data_collection_selector.pxi"
include "_selection_routines/ellipsoid_selector.pxi"
include "_selection_routines/grid_selector.pxi"
include "_selection_routines/octree_subset_selector.pxi"
include "_selection_routines/indexed_octree_subset_selector.pxi"
include "_selection_routines/always_selector.pxi"
include "_selection_routines/compose_selector.pxi"
include "_selection_routines/halo_particles_selector.pxi"
include "_selection_routines/boolean_selectors.pxi"



import numpy as np

from yt.utilities.lib.pixelization_routines import pixelize_cartesian, pixelize_cylinder

from .coordinate_handler import (
    CoordinateHandler,
    _get_coord_fields,
    _setup_dummy_cartesian_coords_and_widths,
)


class GeographicCoordinateHandler(CoordinateHandler):
    radial_axis = "altitude"
    name = "geographic"

    def __init__(self, ds, ordering=None):
        if ordering is None:
            ordering = ("latitude", "longitude", self.radial_axis)
        super().__init__(ds, ordering)
        self.image_units = {}
        self.image_units[self.axis_id["latitude"]] = (None, None)
        self.image_units[self.axis_id["longitude"]] = (None, None)
        self.image_units[self.axis_id[self.radial_axis]] = ("deg", "deg")

    def setup_fields(self, registry):
        # Missing implementation for x, y and z coordinates.
        _setup_dummy_cartesian_coords_and_widths(registry, axes=("x", "y", "z"))

        f1, f2 = _get_coord_fields(self.axis_id["latitude"], "")
        registry.add_field(
            ("index", "dlatitude"),
            sampling_type="cell",
            function=f1,
            display_field=False,
            units="",
        )

        registry.add_field(
            ("index", "latitude"),
            sampling_type="cell",
            function=f2,
            display_field=False,
            units="",
        )

        f1, f2 = _get_coord_fields(self.axis_id["longitude"], "")
        registry.add_field(
            ("index", "dlongitude"),
            sampling_type="cell",
            function=f1,
            display_field=False,
            units="",
        )

        registry.add_field(
            ("index", "longitude"),
            sampling_type="cell",
            function=f2,
            display_field=False,
            units="",
        )

        f1, f2 = _get_coord_fields(self.axis_id[self.radial_axis])
        registry.add_field(
            ("index", f"d{self.radial_axis}"),
            sampling_type="cell",
            function=f1,
            display_field=False,
            units="code_length",
        )

        registry.add_field(
            ("index", self.radial_axis),
            sampling_type="cell",
            function=f2,
            display_field=False,
            units="code_length",
        )

        def _SphericalVolume(field, data):
            # We can use the transformed coordinates here.
            # Here we compute the spherical volume element exactly
            r = data["index", "r"]
            dr = data["index", "dr"]
            theta = data["index", "theta"]
            dtheta = data["index", "dtheta"]
            vol = ((r + 0.5 * dr) ** 3 - (r - 0.5 * dr) ** 3) / 3.0
            vol *= np.cos(theta - 0.5 * dtheta) - np.cos(theta + 0.5 * dtheta)
            vol *= data["index", "dphi"]
            return vol

        registry.add_field(
            ("index", "cell_volume"),
            sampling_type="cell",
            function=_SphericalVolume,
            units="code_length**3",
        )
        registry.alias(("index", "volume"), ("index", "cell_volume"))

        def _path_radial_axis(field, data):
            return data["index", f"d{self.radial_axis}"]

        registry.add_field(
            ("index", f"path_element_{self.radial_axis}"),
            sampling_type="cell",
            function=_path_radial_axis,
            units="code_length",
        )

        def _path_latitude(field, data):
            # We use r here explicitly
            return data["index", "r"] * data["index", "dlatitude"] * np.pi / 180.0

        registry.add_field(
            ("index", "path_element_latitude"),
            sampling_type="cell",
            function=_path_latitude,
            units="code_length",
        )

        def _path_longitude(field, data):
            # We use r here explicitly
            return (
                data["index", "r"]
                * data["index", "dlongitude"]
                * np.pi
                / 180.0
                * np.sin((data["index", "latitude"] + 90.0) * np.pi / 180.0)
            )

        registry.add_field(
            ("index", "path_element_longitude"),
            sampling_type="cell",
            function=_path_longitude,
            units="code_length",
        )

        def _latitude_to_theta(field, data):
            # latitude runs from -90 to 90
            return (data[("index", "latitude")] + 90) * np.pi / 180.0

        registry.add_field(
            ("index", "theta"),
            sampling_type="cell",
            function=_latitude_to_theta,
            units="",
        )

        def _dlatitude_to_dtheta(field, data):
            return data[("index", "dlatitude")] * np.pi / 180.0

        registry.add_field(
            ("index", "dtheta"),
            sampling_type="cell",
            function=_dlatitude_to_dtheta,
            units="",
        )

        def _longitude_to_phi(field, data):
            # longitude runs from -180 to 180
            return (data[("index", "longitude")] + 180) * np.pi / 180.0

        registry.add_field(
            ("index", "phi"), sampling_type="cell", function=_longitude_to_phi, units=""
        )

        def _dlongitude_to_dphi(field, data):
            return data[("index", "dlongitude")] * np.pi / 180.0

        registry.add_field(
            ("index", "dphi"),
            sampling_type="cell",
            function=_dlongitude_to_dphi,
            units="",
        )

        self._setup_radial_fields(registry)

    def _setup_radial_fields(self, registry):
        # This stays here because we don't want to risk the field detector not
        # properly getting the data_source, etc.
        def _altitude_to_radius(field, data):
            surface_height = data.get_field_parameter("surface_height")
            if surface_height is None:
                if hasattr(data.ds, "surface_height"):
                    surface_height = data.ds.surface_height
                else:
                    surface_height = data.ds.quan(0.0, "code_length")
            return data[("index", "altitude")] + surface_height

        registry.add_field(
            ("index", "r"),
            sampling_type="cell",
            function=_altitude_to_radius,
            units="code_length",
        )
        registry.alias(("index", "dr"), ("index", "daltitude"))

    def _retrieve_radial_offset(self, data_source=None):
        # This returns the factor by which the radial field is multiplied and
        # the scalar its offset by.  Typically the "factor" will *only* be
        # either 1.0 or -1.0.  The order will be factor * r + offset.
        # Altitude is the radius from the central zone minus the radius of the
        # surface.  Depth to radius is negative value of depth plus the
        # outermost radius.
        surface_height = None
        if data_source is not None:
            surface_height = data_source.get_field_parameter("surface_height")
        if surface_height is None:
            if hasattr(self.ds, "surface_height"):
                surface_height = self.ds.surface_height
            else:
                surface_height = self.ds.quan(0.0, "code_length")
        return surface_height, 1.0

    def pixelize(
        self, dimension, data_source, field, bounds, size, antialias=True, periodic=True
    ):
        if self.axis_name[dimension] in ("latitude", "longitude"):
            return self._cyl_pixelize(
                data_source, field, bounds, size, antialias, dimension
            )
        elif self.axis_name[dimension] == self.radial_axis:
            return self._ortho_pixelize(
                data_source, field, bounds, size, antialias, dimension, periodic
            )
        else:
            raise NotImplementedError

    def pixelize_line(self, field, start_point, end_point, npoints):
        raise NotImplementedError

    def _ortho_pixelize(
        self, data_source, field, bounds, size, antialias, dimension, periodic
    ):
        period = self.period[:2].copy()
        period[0] = self.period[self.x_axis[dimension]]
        period[1] = self.period[self.y_axis[dimension]]
        if hasattr(period, "in_units"):
            period = period.in_units("code_length").d

        # For a radial axis, px will correspond to longitude and py will
        # correspond to latitude.
        px = data_source["px"]
        pdx = data_source["pdx"]
        py = data_source["py"]
        pdy = data_source["pdy"]
        buff = np.full((size[1], size[0]), np.nan, dtype="float64")
        pixelize_cartesian(
            buff,
            px,
            py,
            pdx,
            pdy,
            data_source[field],
            bounds,
            int(antialias),
            period,
            int(periodic),
        )
        return buff

    def _cyl_pixelize(self, data_source, field, bounds, size, antialias, dimension):
        offset, factor = self._retrieve_radial_offset(data_source)
        r = factor * data_source["py"] + offset
        # Because of the axis-ordering, dimensions 0 and 1 both have r as py
        # and the angular coordinate as px.  But we need to figure out how to
        # convert our coordinate back to an actual angle, based on which
        # dimension we're in.
        pdx = data_source["pdx"].d * np.pi / 180
        if self.axis_name[self.x_axis[dimension]] == "latitude":
            px = (data_source["px"].d + 90) * np.pi / 180
            do_transpose = True
        elif self.axis_name[self.x_axis[dimension]] == "longitude":
            px = (data_source["px"].d + 180) * np.pi / 180
            do_transpose = False
        else:
            # We should never get here!
            raise NotImplementedError
        buff = np.full((size[1], size[0]), np.nan, dtype="f8")
        pixelize_cylinder(
            buff, r, data_source["pdy"], px, pdx, data_source[field], bounds
        )
        if do_transpose:
            buff = buff.transpose()
        return buff

    def convert_from_cartesian(self, coord):
        raise NotImplementedError

    def convert_to_cartesian(self, coord):
        offset, factor = self._retrieve_radial_offset()
        if isinstance(coord, np.ndarray) and len(coord.shape) > 1:
            rad = self.axis_id[self.radial_axis]
            lon = self.axis_id["longitude"]
            lat = self.axis_id["latitude"]
            r = factor * coord[:, rad] + offset
            theta = coord[:, lon] * np.pi / 180
            phi = coord[:, lat] * np.pi / 180
            nc = np.zeros_like(coord)
            # r, theta, phi
            nc[:, lat] = np.cos(phi) * np.sin(theta) * r
            nc[:, lon] = np.sin(phi) * np.sin(theta) * r
            nc[:, rad] = np.cos(theta) * r
        else:
            a, b, c = coord
            theta = b * np.pi / 180
            phi = a * np.pi / 180
            r = factor * c + offset
            nc = (
                np.cos(phi) * np.sin(theta) * r,
                np.sin(phi) * np.sin(theta) * r,
                np.cos(theta) * r,
            )
        return nc

    def convert_to_cylindrical(self, coord):
        raise NotImplementedError

    def convert_from_cylindrical(self, coord):
        raise NotImplementedError

    def convert_to_spherical(self, coord):
        raise NotImplementedError

    def convert_from_spherical(self, coord):
        raise NotImplementedError

    _image_axis_name = None

    @property
    def image_axis_name(self):
        if self._image_axis_name is not None:
            return self._image_axis_name
        # This is the x and y axes labels that get displayed.  For
        # non-Cartesian coordinates, we usually want to override these for
        # Cartesian coordinates, since we transform them.
        rv = {
            self.axis_id["latitude"]: (
                "x / \\sin(\\mathrm{latitude})",
                "y / \\sin(\\mathrm{latitude})",
            ),
            self.axis_id["longitude"]: ("R", "z"),
            self.axis_id[self.radial_axis]: ("longitude", "latitude"),
        }
        for i in list(rv.keys()):
            rv[self.axis_name[i]] = rv[i]
            rv[self.axis_name[i].capitalize()] = rv[i]
        self._image_axis_name = rv
        return rv

    _x_pairs = (
        ("latitude", "longitude"),
        ("longitude", "latitude"),
        ("altitude", "longitude"),
    )

    _y_pairs = (
        ("latitude", "altitude"),
        ("longitude", "altitude"),
        ("altitude", "latitude"),
    )

    _data_projection = None

    @property
    def data_projection(self):
        # this will control the default projection to use when displaying data
        if self._data_projection is not None:
            return self._data_projection
        dpj = {}
        for ax in self.axis_order:
            if ax == self.radial_axis:
                dpj[ax] = "Mollweide"
            else:
                dpj[ax] = None
        self._data_projection = dpj
        return dpj

    _data_transform = None

    @property
    def data_transform(self):
        # this is the coordinate system on which the data is defined (the crs).
        if self._data_transform is not None:
            return self._data_transform
        dtx = {}
        for ax in self.axis_order:
            if ax == self.radial_axis:
                dtx[ax] = "PlateCarree"
            else:
                dtx[ax] = None
        self._data_transform = dtx
        return dtx

    @property
    def period(self):
        return self.ds.domain_width

    def sanitize_center(self, center, axis):
        center, display_center = super().sanitize_center(center, axis)
        name = self.axis_name[axis]
        if name == self.radial_axis:
            display_center = center
        elif name == "latitude":
            display_center = (
                0.0 * display_center[0],
                0.0 * display_center[1],
                0.0 * display_center[2],
            )
        elif name == "longitude":
            ri = self.axis_id[self.radial_axis]
            c = (self.ds.domain_right_edge[ri] + self.ds.domain_left_edge[ri]) / 2.0
            display_center = [
                0.0 * display_center[0],
                0.0 * display_center[1],
                0.0 * display_center[2],
            ]
            display_center[self.axis_id["latitude"]] = c
        return center, display_center

    def sanitize_width(self, axis, width, depth):
        name = self.axis_name[axis]
        if width is not None:
            width = super().sanitize_width(axis, width, depth)
        elif name == self.radial_axis:
            rax = self.radial_axis
            width = [
                self.ds.domain_width[self.x_axis[rax]],
                self.ds.domain_width[self.y_axis[rax]],
            ]
        elif name == "latitude":
            ri = self.axis_id[self.radial_axis]
            # Remember, in spherical coordinates when we cut in theta,
            # we create a conic section
            width = [2.0 * self.ds.domain_width[ri], 2.0 * self.ds.domain_width[ri]]
        elif name == "longitude":
            ri = self.axis_id[self.radial_axis]
            width = [self.ds.domain_width[ri], 2.0 * self.ds.domain_width[ri]]
        return width


class InternalGeographicCoordinateHandler(GeographicCoordinateHandler):
    radial_axis = "depth"
    name = "internal_geographic"

    def _setup_radial_fields(self, registry):
        # Altitude is the radius from the central zone minus the radius of the
        # surface.
        def _depth_to_radius(field, data):
            outer_radius = data.get_field_parameter("outer_radius")
            if outer_radius is None:
                if hasattr(data.ds, "outer_radius"):
                    outer_radius = data.ds.outer_radius
                else:
                    # Otherwise, we assume that the depth goes to full depth,
                    # so we can look at the domain right edge in depth.
                    rax = self.axis_id[self.radial_axis]
                    outer_radius = data.ds.domain_right_edge[rax]
            return -1.0 * data[("index", "depth")] + outer_radius

        registry.add_field(
            ("index", "r"),
            sampling_type="cell",
            function=_depth_to_radius,
            units="code_length",
        )
        registry.alias(("index", "dr"), ("index", "ddepth"))

    def _retrieve_radial_offset(self, data_source=None):
        # Depth means switching sign and adding to full radius
        outer_radius = None
        if data_source is not None:
            outer_radius = data_source.get_field_parameter("outer_radius")
        if outer_radius is None:
            if hasattr(self.ds, "outer_radius"):
                outer_radius = self.ds.outer_radius
            else:
                # Otherwise, we assume that the depth goes to full depth,
                # so we can look at the domain right edge in depth.
                rax = self.axis_id[self.radial_axis]
                outer_radius = self.ds.domain_right_edge[rax]
        return outer_radius, -1.0

    _x_pairs = (
        ("latitude", "longitude"),
        ("longitude", "latitude"),
        ("depth", "longitude"),
    )

    _y_pairs = (("latitude", "depth"), ("longitude", "depth"), ("depth", "latitude"))

    def sanitize_center(self, center, axis):
        center, display_center = super(
            GeographicCoordinateHandler, self
        ).sanitize_center(center, axis)
        name = self.axis_name[axis]
        if name == self.radial_axis:
            display_center = center
        elif name == "latitude":
            display_center = (
                0.0 * display_center[0],
                0.0 * display_center[1],
                0.0 * display_center[2],
            )
        elif name == "longitude":
            ri = self.axis_id[self.radial_axis]
            offset, factor = self._retrieve_radial_offset()
            outermost = factor * self.ds.domain_left_edge[ri] + offset
            display_center = [
                0.0 * display_center[0],
                0.0 * display_center[1],
                0.0 * display_center[2],
            ]
            display_center[self.axis_id["latitude"]] = outermost / 2.0
        return center, display_center

    def sanitize_width(self, axis, width, depth):
        name = self.axis_name[axis]
        if width is not None:
            width = super(GeographicCoordinateHandler, self).sanitize_width(
                axis, width, depth
            )
        elif name == self.radial_axis:
            rax = self.radial_axis
            width = [
                self.ds.domain_width[self.x_axis[rax]],
                self.ds.domain_width[self.y_axis[rax]],
            ]
        elif name == "latitude":
            ri = self.axis_id[self.radial_axis]
            # Remember, in spherical coordinates when we cut in theta,
            # we create a conic section
            offset, factor = self._retrieve_radial_offset()
            outermost = factor * self.ds.domain_left_edge[ri] + offset
            width = [2.0 * outermost, 2.0 * outermost]
        elif name == "longitude":
            ri = self.axis_id[self.radial_axis]
            offset, factor = self._retrieve_radial_offset()
            outermost = factor * self.ds.domain_left_edge[ri] + offset
            width = [outermost, 2.0 * outermost]
        return width



from yt.fields.field_info_container import FieldInfoContainer


class SDFFieldInfo(FieldInfoContainer):
    known_other_fields = ()

    known_particle_fields = ()
    _mass_field = None

    def __init__(self, ds, field_list):
        if "mass" in field_list:
            self.known_particle_fields.append(
                ("mass", "code_mass", ["particle_mass"], None)
            )
        possible_masses = ["mass", "m200b", "mvir"]
        mnf = "mass"
        for mn in possible_masses:
            if mn in ds.sdf_container.keys():
                mnf = self._mass_field = mn
                break

        idf = ds._field_map.get("particle_index", "ident")
        xf = ds._field_map.get("particle_position_x", "x")
        yf = ds._field_map.get("particle_position_y", "y")
        zf = ds._field_map.get("particle_position_z", "z")
        vxf = ds._field_map.get("particle_velocity_x", "vx")
        vyf = ds._field_map.get("particle_velocity_z", "vy")
        vzf = ds._field_map.get("particle_velocity_z", "vz")

        self.known_particle_fields = (
            (idf, ("dimensionless", ["particle_index"], None)),
            (xf, ("code_length", ["particle_position_x"], None)),
            (yf, ("code_length", ["particle_position_y"], None)),
            (zf, ("code_length", ["particle_position_z"], None)),
            (vxf, ("code_velocity", ["particle_velocity_x"], None)),
            (vyf, ("code_velocity", ["particle_velocity_y"], None)),
            (vzf, ("code_velocity", ["particle_velocity_z"], None)),
            (mnf, ("code_mass", ["particle_mass"], None)),
        )
        super().__init__(ds, field_list)



import os.path

from numpy.testing import assert_equal

from yt.frontends.ahf.api import AHFHalosDataset
from yt.testing import ParticleSelectionComparison, requires_file
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    data_dir_load,
    requires_ds,
)

_fields = (
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_position_z"),
    ("all", "particle_mass"),
)

ahf_halos = "ahf_halos/snap_N64L16_135.parameter"


def load(filename):
    return data_dir_load(filename, kwargs={"hubble_constant": 0.7})


@requires_ds(ahf_halos)
def test_fields_ahf_halos():
    ds = load(ahf_halos)
    assert_equal(str(ds), os.path.basename(ahf_halos))
    for field in _fields:
        yield FieldValuesTest(ds, field, particle_type=True)


@requires_file(ahf_halos)
def test_AHFHalosDataset():
    ds = load(ahf_halos)
    assert isinstance(ds, AHFHalosDataset)
    ad = ds.all_data()
    ad[("all", "particle_mass")]
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()



import numpy as np
from matplotlib.colors import LogNorm

import yt
from yt.visualization.api import get_multi_plot

fn = "Enzo_64/RD0006/RedshiftOutput0006"  # dataset to load

# load data and get center value and center location as maximum density location
ds = yt.load(fn)
v, c = ds.find_max(("gas", "density"))

# set up our Fixed Resolution Buffer parameters: a width, resolution, and center
width = (1.0, "unitary")
res = [1000, 1000]
#  get_multi_plot returns a containing figure, a list-of-lists of axes
#   into which we can place plots, and some axes that we'll put
#   colorbars.

#  it accepts: # of x-axis plots, # of y-axis plots, and how the
#  colorbars are oriented (this also determines where they go: below
#  in the case of 'horizontal', on the right in the case of
#  'vertical'), bw is the base-width in inches (4 is about right for
#  most cases)

orient = "horizontal"
fig, axes, colorbars = get_multi_plot(2, 3, colorbar=orient, bw=6)

# Now we follow the method of "multi_plot.py" but we're going to iterate
# over the columns, which will become axes of slicing.
plots = []
for ax in range(3):
    sli = ds.slice(ax, c[ax])
    frb = sli.to_frb(width, res)
    den_axis = axes[ax][0]
    temp_axis = axes[ax][1]

    # here, we turn off the axes labels and ticks, but you could
    # customize further.
    for ax in (den_axis, temp_axis):
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)

    # converting our fixed resolution buffers to NDarray so matplotlib can
    # render them
    dens = np.array(frb[("gas", "density")])
    temp = np.array(frb[("gas", "temperature")])

    plots.append(den_axis.imshow(dens, norm=LogNorm()))
    plots[-1].set_clim((5e-32, 1e-29))
    plots[-1].set_cmap("bds_highcontrast")

    plots.append(temp_axis.imshow(temp, norm=LogNorm()))
    plots[-1].set_clim((1e3, 1e8))
    plots[-1].set_cmap("hot")

# Each 'cax' is a colorbar-container, into which we'll put a colorbar.
# the zip command creates triples from each element of the three lists
# .  Note that it cuts off after the shortest iterator is exhausted,
# in this case, titles.
titles = [
    r"$\mathrm{density}\ (\mathrm{g\ cm^{-3}})$",
    r"$\mathrm{temperature}\ (\mathrm{K})$",
]
for p, cax, t in zip(plots, colorbars, titles):
    # Now we make a colorbar, using the 'image' we stored in plots
    # above. note this is what is *returned* by the imshow method of
    # the plots.
    cbar = fig.colorbar(p, cax=cax, orientation=orient)
    cbar.set_label(t)

# And now we're done!
fig.savefig(f"{ds}_3x2.png")



import numpy as np
from numpy.testing import assert_equal

from yt.utilities.lib.misc_utilities import fill_region

NDIM = 32


def test_fill_region():
    for level in range(2):
        rf = 2**level
        output_fields = [
            np.zeros((NDIM * rf, NDIM * rf, NDIM * rf), "float64") for i in range(3)
        ]
        input_fields = [np.empty(NDIM**3, "float64") for i in range(3)]
        v = np.mgrid[
            0.0 : 1.0 : NDIM * 1j, 0.0 : 1.0 : NDIM * 1j, 0.0 : 1.0 : NDIM * 1j
        ]
        input_fields[0][:] = v[0].ravel()
        input_fields[1][:] = v[1].ravel()
        input_fields[2][:] = v[2].ravel()
        left_index = np.zeros(3, "int64")
        ipos = np.empty((NDIM**3, 3), dtype="int64")
        ind = np.indices((NDIM, NDIM, NDIM))
        ipos[:, 0] = ind[0].ravel()
        ipos[:, 1] = ind[1].ravel()
        ipos[:, 2] = ind[2].ravel()
        ires = np.zeros(NDIM * NDIM * NDIM, "int64")
        ddims = np.array([NDIM, NDIM, NDIM], dtype="int64") * rf
        fill_region(
            input_fields,
            output_fields,
            level,
            left_index,
            ipos,
            ires,
            ddims,
            np.array([2, 2, 2], dtype="i8"),
        )
        for r in range(level + 1):
            for o, i in zip(output_fields, v):
                assert_equal(o[r::rf, r::rf, r::rf], i)



# Some tests for the Cartesian coordinates handler

import numpy as np
from numpy.testing import assert_equal

from yt.testing import fake_amr_ds

# Our canonical tests are that we can access all of our fields and we can
# compute our volume correctly.


def test_cartesian_coordinates():
    # We're going to load up a simple AMR grid and check its volume
    # calculations and path length calculations.
    ds = fake_amr_ds()
    axes = sorted(set(ds.coordinates.axis_name.values()))
    for i, axis in enumerate(axes):
        dd = ds.all_data()
        fi = ("index", axis)
        fd = ("index", f"d{axis}")
        fp = ("index", f"path_element_{axis}")
        ma = np.argmax(dd[fi])
        assert_equal(dd[fi][ma] + dd[fd][ma] / 2.0, ds.domain_right_edge[i])
        mi = np.argmin(dd[fi])
        assert_equal(dd[fi][mi] - dd[fd][mi] / 2.0, ds.domain_left_edge[i])
        assert_equal(dd[fd].min(), ds.index.get_smallest_dx())
        assert_equal(dd[fd].max(), (ds.domain_width / ds.domain_dimensions)[i])
        assert_equal(dd[fd], dd[fp])
    assert_equal(
        dd[("index", "cell_volume")].sum(dtype="float64"), ds.domain_width.prod()
    )



from yt.testing import assert_allclose_units, fake_amr_ds


def test_icoords_to_ires():
    for geometry in ("cartesian", "spherical", "cylindrical"):
        ds = fake_amr_ds(geometry=geometry)

        dd = ds.r[:]
        icoords = dd.icoords
        ires = dd.ires
        fcoords, fwidth = ds.index._icoords_to_fcoords(icoords, ires)
        assert_allclose_units(fcoords, dd.fcoords, rtol=1e-14)
        assert_allclose_units(fwidth, dd.fwidth, rtol=1e-14)

        fcoords_xz, fwidth_xz = ds.index._icoords_to_fcoords(
            icoords[:, (0, 2)], ires, axes=(0, 2)
        )
        assert_allclose_units(fcoords_xz[:, 0], dd.fcoords[:, 0])
        assert_allclose_units(fcoords_xz[:, 1], dd.fcoords[:, 2])
        assert_allclose_units(fwidth_xz[:, 0], dd.fwidth[:, 0])
        assert_allclose_units(fwidth_xz[:, 1], dd.fwidth[:, 2])



import unittest

import numpy as np

import yt
from yt.testing import fake_amr_ds, requires_module
from yt.visualization.geo_plot_utils import get_mpl_transform, transform_list


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class TestGeoProjections(unittest.TestCase):
    @requires_module("cartopy")
    def setUp(self):
        self.ds = fake_amr_ds(geometry="geographic")
        # switch off the log plot to avoid some unrelated matplotlib issues
        f = self.ds._get_field_info(("stream", "Density"))
        f.take_log = False

    @requires_module("cartopy")
    def tearDown(self):
        del self.ds

    @requires_module("cartopy")
    def test_geo_projection_setup(self):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        axis = "altitude"
        self.slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")

        assert isinstance(self.slc._projection, cartopy.crs.Mollweide)
        assert isinstance(self.slc._transform, cartopy.crs.PlateCarree)
        assert self.ds.coordinates.data_projection[axis] == "Mollweide"
        assert self.ds.coordinates.data_transform[axis] == "PlateCarree"
        assert isinstance(
            self.slc._projection,
            type(self.slc.plots[("stream", "Density")].axes.projection),
        )

    @requires_module("cartopy")
    def test_geo_projections(self):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        self.slc = yt.SlicePlot(
            self.ds, "altitude", ("stream", "Density"), origin="native"
        )

        for transform in transform_list:
            if transform == "UTM":
                # this requires special arguments so let's skip it
                continue
            if transform == "OSNI":
                # avoid crashes, see https://github.com/SciTools/cartopy/issues/1177
                continue
            self.slc.set_mpl_projection(transform)
            proj_type = type(get_mpl_transform(transform))

            assert isinstance(self.slc._projection, proj_type)
            assert isinstance(self.slc._transform, cartopy.crs.PlateCarree)
            assert isinstance(
                self.slc.plots[("stream", "Density")].axes.projection, proj_type
            )

    @requires_module("cartopy")
    def test_projection_object(self):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        shortlist = ["Orthographic", "PlateCarree", "Mollweide"]

        for transform in shortlist:
            projection = get_mpl_transform(transform)
            proj_type = type(projection)
            self.slc = yt.SlicePlot(
                self.ds, "altitude", ("stream", "Density"), origin="native"
            )
            self.slc.set_mpl_projection(projection)

            assert isinstance(self.slc._projection, proj_type)
            assert isinstance(self.slc._transform, cartopy.crs.PlateCarree)
            assert isinstance(
                self.slc.plots[("stream", "Density")].axes.projection, proj_type
            )

    @requires_module("cartopy")
    def test_nondefault_transform(self):
        from yt.utilities.on_demand_imports import _cartopy as cartopy

        axis = "altitude"
        # Note: The Miller transform has an extent of approx. +/- 180 in x,
        # +/-132 in y (in Miller, x is longitude, y is a factor of latitude).
        # So by changing the projection in this way, the dataset goes from
        # covering the whole globe (+/- 90 latitude), to covering part of the
        # globe (+/-72 latitude). Totally fine for testing that the code runs,
        # but good to be aware of!
        self.ds.coordinates.data_transform[axis] = "Miller"
        self.slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")

        shortlist = ["Orthographic", "PlateCarree", "Mollweide"]

        for transform in shortlist:
            self.slc.set_mpl_projection(transform)
            proj_type = type(get_mpl_transform(transform))

            assert isinstance(self.slc._projection, proj_type)
            assert isinstance(self.slc._transform, cartopy.crs.Miller)
            assert self.ds.coordinates.data_projection[axis] == "Mollweide"
            assert self.ds.coordinates.data_transform[axis] == "Miller"
            assert isinstance(
                self.slc.plots[("stream", "Density")].axes.projection, proj_type
            )

    @requires_module("cartopy")
    def test_extent(self):
        # checks that the axis extent is narrowed when doing a subselection
        axis = "altitude"
        slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")
        ax = slc.plots[("stream", "Density")].axes
        full_extent = np.abs(ax.get_extent())

        slc = yt.SlicePlot(
            self.ds, axis, ("stream", "Density"), origin="native", width=(80.0, 50.0)
        )
        ax = slc.plots[("stream", "Density")].axes
        extent = np.abs(ax.get_extent())
        assert np.all(extent < full_extent)


class TestNonGeoProjections(unittest.TestCase):
    def setUp(self):
        self.ds = fake_amr_ds()

    def tearDown(self):
        del self.ds
        del self.slc

    def test_projection_setup(self):
        axis = "x"
        self.slc = yt.SlicePlot(self.ds, axis, ("stream", "Density"), origin="native")

        assert self.ds.coordinates.data_projection[axis] is None
        assert self.ds.coordinates.data_transform[axis] is None
        assert self.slc._projection is None
        assert self.slc._transform is None



# these imports are very expensive so we delay them to until they are requested
from matplotlib.backend_bases import FigureCanvasBase
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_pdf import FigureCanvasPdf
from matplotlib.backends.backend_ps import FigureCanvasPS
from matplotlib.backends.backend_svg import FigureCanvasSVG



import os
import shutil
import tempfile

import numpy as np
from numpy.testing import assert_allclose, assert_equal

from yt.loaders import load
from yt.testing import fake_random_ds, requires_file, requires_module
from yt.utilities.on_demand_imports import _astropy
from yt.visualization.fits_image import (
    FITSImageData,
    FITSOffAxisProjection,
    FITSOffAxisSlice,
    FITSParticleOffAxisProjection,
    FITSParticleProjection,
    FITSProjection,
    FITSSlice,
    assert_same_wcs,
)
from yt.visualization.volume_rendering.off_axis_projection import off_axis_projection


@requires_module("astropy")
def test_fits_image():
    curdir = os.getcwd()
    tmpdir = tempfile.mkdtemp()
    os.chdir(tmpdir)

    fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
    units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
    ds = fake_random_ds(
        64, fields=fields, units=units, nprocs=16, length_unit=100.0, particles=10000
    )

    prj = ds.proj(("gas", "density"), 2)
    prj_frb = prj.to_frb((0.5, "unitary"), 128)

    fid1 = prj_frb.to_fits_data(
        fields=[("gas", "density"), ("gas", "temperature")], length_unit="cm"
    )
    fits_prj = FITSProjection(
        ds,
        "z",
        [ds.fields.gas.density, ("gas", "temperature")],
        image_res=128,
        width=(0.5, "unitary"),
    )

    assert_equal(fid1["density"].data, fits_prj["density"].data)
    assert_equal(fid1["temperature"].data, fits_prj["temperature"].data)

    fid1.writeto("fid1.fits", overwrite=True)
    new_fid1 = FITSImageData.from_file("fid1.fits")

    assert_equal(fid1["density"].data, new_fid1["density"].data)
    assert_equal(fid1["temperature"].data, new_fid1["temperature"].data)
    assert_equal(fid1.length_unit, new_fid1.length_unit)
    assert_equal(fid1.time_unit, new_fid1.time_unit)
    assert_equal(fid1.mass_unit, new_fid1.mass_unit)
    assert_equal(fid1.velocity_unit, new_fid1.velocity_unit)
    assert_equal(fid1.magnetic_unit, new_fid1.magnetic_unit)
    assert_equal(fid1.current_time, new_fid1.current_time)

    ds2 = load("fid1.fits")
    ds2.index

    assert ("fits", "density") in ds2.field_list
    assert ("fits", "temperature") in ds2.field_list

    dw_cm = ds2.domain_width.in_units("cm")

    assert dw_cm[0].v == 50.0
    assert dw_cm[1].v == 50.0

    slc = ds.slice(2, 0.5)
    slc_frb = slc.to_frb((0.5, "unitary"), 128)

    fid2 = slc_frb.to_fits_data(
        fields=[("gas", "density"), ("gas", "temperature")], length_unit="cm"
    )
    fits_slc = FITSSlice(
        ds,
        "z",
        [("gas", "density"), ("gas", "temperature")],
        image_res=128,
        width=(0.5, "unitary"),
    )

    assert_equal(fid2["density"].data, fits_slc["density"].data)
    assert_equal(fid2["temperature"].data, fits_slc["temperature"].data)

    fits_slc2 = FITSSlice(
        ds,
        "z",
        [("gas", "density"), ("gas", "temperature")],
        image_res=128,
        width=(0.5, "unitary"),
        origin="image",
    )

    assert_equal(fits_slc2["density"].data, fits_slc["density"].data)
    assert_equal(fits_slc2["temperature"].data, fits_slc["temperature"].data)
    assert fits_slc.wcs.wcs.crval[0] != 0.0
    assert fits_slc.wcs.wcs.crval[1] != 0.0
    assert fits_slc2.wcs.wcs.crval[0] == 0.0
    assert fits_slc2.wcs.wcs.crval[1] == 0.0

    dens_img = fid2.pop("density")
    temp_img = fid2.pop("temperature")

    combined_fid = FITSImageData.from_images([dens_img, temp_img])
    assert_equal(combined_fid.length_unit, dens_img.length_unit)
    assert_equal(combined_fid.time_unit, dens_img.time_unit)
    assert_equal(combined_fid.mass_unit, dens_img.mass_unit)
    assert_equal(combined_fid.velocity_unit, dens_img.velocity_unit)
    assert_equal(combined_fid.magnetic_unit, dens_img.magnetic_unit)
    assert_equal(combined_fid.current_time, dens_img.current_time)

    # Make sure that we can combine FITSImageData instances with more
    # than one image each
    combined_fid2 = FITSImageData.from_images([combined_fid, combined_fid])
    # Writing the FITS file ensures that we have assembled the images
    # together correctly
    combined_fid2.writeto("combined.fits", overwrite=True)

    cut = ds.cutting([0.1, 0.2, -0.9], [0.5, 0.42, 0.6])
    cut_frb = cut.to_frb((0.5, "unitary"), 128)

    fid3 = cut_frb.to_fits_data(
        fields=[("gas", "density"), ds.fields.gas.temperature], length_unit="cm"
    )
    fits_cut = FITSOffAxisSlice(
        ds,
        [0.1, 0.2, -0.9],
        [("gas", "density"), ("gas", "temperature")],
        image_res=128,
        center=[0.5, 0.42, 0.6],
        width=(0.5, "unitary"),
    )

    assert_equal(fid3["density"].data, fits_cut["density"].data)
    assert_equal(fid3["temperature"].data, fits_cut["temperature"].data)

    fid3.create_sky_wcs([30.0, 45.0], (1.0, "arcsec/kpc"))
    fid3.writeto("fid3.fits", overwrite=True)
    new_fid3 = FITSImageData.from_file("fid3.fits")
    assert_same_wcs(fid3.wcs, new_fid3.wcs)
    assert new_fid3.wcs.wcs.cunit[0] == "deg"
    assert new_fid3.wcs.wcs.cunit[1] == "deg"
    assert new_fid3.wcs.wcs.ctype[0] == "RA---TAN"
    assert new_fid3.wcs.wcs.ctype[1] == "DEC--TAN"
    assert new_fid3.wcsa.wcs.cunit[0] == "cm"
    assert new_fid3.wcsa.wcs.cunit[1] == "cm"
    assert new_fid3.wcsa.wcs.ctype[0] == "linear"
    assert new_fid3.wcsa.wcs.ctype[1] == "linear"

    buf = off_axis_projection(
        ds, ds.domain_center, [0.1, 0.2, -0.9], 0.5, 128, ("gas", "density")
    ).swapaxes(0, 1)
    fid4 = FITSImageData(buf, fields=[("gas", "density")], width=100.0)
    fits_oap = FITSOffAxisProjection(
        ds,
        [0.1, 0.2, -0.9],
        ("gas", "density"),
        width=(0.5, "unitary"),
        image_res=128,
        depth=(0.5, "unitary"),
    )

    assert_equal(fid4["density"].data, fits_oap["density"].data)

    fid4.create_sky_wcs([30.0, 45.0], (1.0, "arcsec/kpc"), replace_old_wcs=False)
    assert fid4.wcs.wcs.cunit[0] == "cm"
    assert fid4.wcs.wcs.cunit[1] == "cm"
    assert fid4.wcs.wcs.ctype[0] == "linear"
    assert fid4.wcs.wcs.ctype[1] == "linear"
    assert fid4.wcsa.wcs.cunit[0] == "deg"
    assert fid4.wcsa.wcs.cunit[1] == "deg"
    assert fid4.wcsa.wcs.ctype[0] == "RA---TAN"
    assert fid4.wcsa.wcs.ctype[1] == "DEC--TAN"

    cvg = ds.covering_grid(
        ds.index.max_level,
        [0.25, 0.25, 0.25],
        [32, 32, 32],
        fields=[("gas", "density"), ("gas", "temperature")],
    )
    fid5 = cvg.to_fits_data(fields=[("gas", "density"), ("gas", "temperature")])
    assert fid5.dimensionality == 3

    fid5.update_header("density", "time", 0.1)
    fid5.update_header("all", "units", "cgs")

    assert fid5["density"].header["time"] == 0.1
    assert fid5["temperature"].header["units"] == "cgs"
    assert fid5["density"].header["units"] == "cgs"

    fid6 = FITSImageData.from_images(fid5)

    fid5.change_image_name("density", "mass_per_volume")
    assert fid5["mass_per_volume"].name == "mass_per_volume"
    assert fid5["mass_per_volume"].header["BTYPE"] == "mass_per_volume"
    assert "mass_per_volume" in fid5.fields
    assert "mass_per_volume" in fid5.field_units
    assert "density" not in fid5.fields
    assert "density" not in fid5.field_units

    assert "density" in fid6.fields
    assert_equal(fid6["density"].data, fid5["mass_per_volume"].data)

    fid7 = FITSImageData.from_images(fid4)
    fid7.convolve("density", (3.0, "cm"))

    sigma = 3.0 / fid7.wcs.wcs.cdelt[0]
    kernel = _astropy.conv.Gaussian2DKernel(x_stddev=sigma)
    data_conv = _astropy.conv.convolve(fid4["density"].data.d, kernel)
    assert_allclose(data_conv, fid7["density"].data.d)

    pfid = FITSParticleProjection(ds, "x", ("io", "particle_mass"))
    assert pfid["particle_mass"].name == "particle_mass"
    assert pfid["particle_mass"].header["BTYPE"] == "particle_mass"
    assert pfid["particle_mass"].units == "g"

    pofid = FITSParticleOffAxisProjection(ds, [1, 1, 1], ("io", "particle_mass"))
    assert pofid["particle_mass"].name == "particle_mass"
    assert pofid["particle_mass"].header["BTYPE"] == "particle_mass"
    assert pofid["particle_mass"].units == "g"

    pdfid = FITSParticleProjection(ds, "x", ("io", "particle_mass"), density=True)
    assert pdfid["particle_mass"].name == "particle_mass"
    assert pdfid["particle_mass"].header["BTYPE"] == "particle_mass"
    assert pdfid["particle_mass"].units == "g/cm**2"

    # Test moments for projections
    def _vysq(field, data):
        return data["gas", "velocity_y"] ** 2

    ds.add_field(("gas", "vysq"), _vysq, sampling_type="cell", units="cm**2/s**2")
    fid8 = FITSProjection(
        ds,
        "y",
        [("gas", "velocity_y"), ("gas", "vysq")],
        moment=1,
        weight_field=("gas", "density"),
    )
    fid9 = FITSProjection(
        ds, "y", ("gas", "velocity_y"), moment=2, weight_field=("gas", "density")
    )
    sigy = np.sqrt(fid8["vysq"].data - fid8["velocity_y"].data ** 2)
    assert_allclose(sigy, fid9["velocity_y_stddev"].data)

    def _vlsq(field, data):
        return data["gas", "velocity_los"] ** 2

    ds.add_field(("gas", "vlsq"), _vlsq, sampling_type="cell", units="cm**2/s**2")
    fid10 = FITSOffAxisProjection(
        ds,
        [1.0, -1.0, 1.0],
        [("gas", "velocity_los"), ("gas", "vlsq")],
        moment=1,
        weight_field=("gas", "density"),
    )
    fid11 = FITSOffAxisProjection(
        ds,
        [1.0, -1.0, 1.0],
        ("gas", "velocity_los"),
        moment=2,
        weight_field=("gas", "density"),
    )
    sigl = np.sqrt(fid10["vlsq"].data - fid10["velocity_los"].data ** 2)
    assert_allclose(sigl, fid11["velocity_los_stddev"].data)

    # We need to manually close all the file descriptors so
    # that windows can delete the folder that contains them.
    ds2.close()
    for fid in (
        fid1,
        fid2,
        fid3,
        fid4,
        fid5,
        fid6,
        fid7,
        fid8,
        fid9,
        new_fid1,
        new_fid3,
        pfid,
        pdfid,
        pofid,
    ):
        fid.close()

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


etc = "enzo_tiny_cosmology/DD0046/DD0046"


@requires_module("astropy")
@requires_file(etc)
def test_fits_cosmo():
    ds = load(etc)

    fid = FITSProjection(ds, "z", ["density"])
    assert fid.wcs.wcs.cunit[0] == "kpc"
    assert fid.wcs.wcs.cunit[1] == "kpc"
    assert ds.hubble_constant == fid.hubble_constant
    assert ds.current_redshift == fid.current_redshift
    assert fid["density"].header["HUBBLE"] == ds.hubble_constant
    assert fid["density"].header["REDSHIFT"] == ds.current_redshift
    fid.close()



import numpy as np

from yt.loaders import load_uniform_grid
from yt.testing import requires_module_pytest
from yt.utilities.on_demand_imports import _h5py as h5py


@requires_module_pytest("h5py")
def test_save_as_data_unit_system(tmp_path):
    # This test checks that the file saved with calling save_as_dataset
    # for a ds with a "code" unit system contains the proper "unit_system_name".
    # It checks the hdf5 file directly rather than using yt.load(), because
    # https://github.com/yt-project/yt/issues/4315 only manifested restarting
    # the python kernel (because the unit registry is state dependent).

    fi = tmp_path / "output_data.h5"
    shp = (4, 4, 4)
    data = {"density": np.random.random(shp)}
    ds = load_uniform_grid(data, shp, unit_system="code")
    assert "code" in ds._unit_system_name

    sp = ds.sphere(ds.domain_center, ds.domain_width[0] / 2.0)
    sp.save_as_dataset(fi)

    with h5py.File(fi, mode="r") as f:
        assert f.attrs["unit_system_name"] == "code"



import os
import weakref
from collections import defaultdict
from typing import Optional, Tuple

import numpy as np

from yt.data_objects.field_data import YTFieldData
from yt.data_objects.index_subobjects.octree_subset import OctreeSubset
from yt.data_objects.static_output import Dataset
from yt.data_objects.unions import ParticleUnion
from yt.frontends.artio import _artio_caller
from yt.frontends.artio._artio_caller import (
    ARTIOSFCRangeHandler,
    artio_fileset,
    artio_is_valid,
)
from yt.frontends.artio.fields import ARTIOFieldInfo
from yt.funcs import mylog, setdefaultattr
from yt.geometry import particle_deposit as particle_deposit
from yt.geometry.geometry_handler import Index, YTDataChunk
from yt.utilities.exceptions import YTParticleDepositionNotImplemented


class ARTIOOctreeSubset(OctreeSubset):
    _domain_offset = 0
    domain_id = -1
    _con_args = ("base_region", "sfc_start", "sfc_end", "oct_handler", "ds")
    _type_name = "octree_subset"
    _num_zones = 2

    def __init__(self, base_region, sfc_start, sfc_end, oct_handler, ds):
        self.field_data = YTFieldData()
        self.field_parameters = {}
        self.sfc_start = sfc_start
        self.sfc_end = sfc_end
        self.oct_handler = oct_handler
        self.ds = ds
        self._last_mask = None
        self._last_selector_id = None
        self._current_particle_type = "all"
        self._current_fluid_type = self.ds.default_fluid_type
        self.base_region = base_region
        self.base_selector = base_region.selector

    @property
    def min_ind(self):
        return self.sfc_start

    @property
    def max_ind(self):
        return self.sfc_end

    def fill(self, fields, selector):
        if len(fields) == 0:
            return []
        handle = self.oct_handler.artio_handle
        field_indices = [
            handle.parameters["grid_variable_labels"].index(f) for (ft, f) in fields
        ]
        cell_count = selector.count_oct_cells(self.oct_handler, self.domain_id)
        self.data_size = cell_count
        levels, cell_inds, file_inds = self.oct_handler.file_index_octs(
            selector, self.domain_id, cell_count
        )
        domain_counts = self.oct_handler.domain_count(selector)
        tr = [np.zeros(cell_count, dtype="float64") for field in fields]
        self.oct_handler.fill_sfc(
            levels, cell_inds, file_inds, domain_counts, field_indices, tr
        )
        tr = dict(zip(fields, tr))
        return tr

    def fill_particles(self, fields):
        if len(fields) == 0:
            return {}
        ptype_indices = self.ds.particle_types
        art_fields = [(ptype_indices.index(ptype), fname) for ptype, fname in fields]
        species_data = self.oct_handler.fill_sfc_particles(art_fields)
        tr = defaultdict(dict)
        # Now we need to sum things up and then fill
        for s, f in fields:
            count = 0
            dt = "float64"  # default
            i = ptype_indices.index(s)
            # No vector fields in ARTIO
            count += species_data[i, f].size
            dt = species_data[i, f].dtype
            tr[s][f] = np.zeros(count, dtype=dt)
            cp = 0
            v = species_data.pop((i, f))
            tr[s][f][cp : cp + v.size] = v
            cp += v.size
        return tr


# We create something of a fake octree here.  This is primarily to enable us to
# reuse code for things like __getitem__ and the like.  We will also create a
# new oct_handler type that is functionally equivalent, except that it will
# only manage the root mesh.
class ARTIORootMeshSubset(ARTIOOctreeSubset):
    _num_zones = 1
    _type_name = "sfc_subset"
    _selector_module = _artio_caller
    domain_id = -1

    def fill(self, fields, selector):
        # We know how big these will be.
        if len(fields) == 0:
            return []
        handle = self.ds._handle
        field_indices = [
            handle.parameters["grid_variable_labels"].index(f) for (ft, f) in fields
        ]
        tr = self.oct_handler.fill_sfc(selector, field_indices)
        self.data_size = tr[0].size
        tr = dict(zip(fields, tr))
        return tr

    def deposit(self, positions, fields=None, method=None, kernel_name="cubic"):
        # Here we perform our particle deposition.
        if fields is None:
            fields = []
        cls = getattr(particle_deposit, f"deposit_{method}", None)
        if cls is None:
            raise YTParticleDepositionNotImplemented(method)
        nz = self.nz
        nvals = (nz, nz, nz, self.ires.size)
        # We allocate number of zones, not number of octs
        op = cls(nvals, kernel_name)
        op.initialize()
        mylog.debug(
            "Depositing %s (%s^3) particles into %s Root Mesh",
            positions.shape[0],
            positions.shape[0] ** 0.3333333,
            nvals[-1],
        )
        pos = np.array(positions, dtype="float64")
        f64 = [np.array(f, dtype="float64") for f in fields]
        self.oct_handler.deposit(op, self.base_selector, pos, f64)
        vals = op.finalize()
        if vals is None:
            return
        return np.asfortranarray(vals)


class ARTIOIndex(Index):
    def __init__(self, ds, dataset_type="artio"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)

        self.max_level = ds.max_level
        self.range_handlers = {}
        self.float_type = np.float64
        super().__init__(ds, dataset_type)

    @property
    def max_range(self):
        return self.dataset.max_range

    def _setup_geometry(self):
        mylog.debug("Initializing Geometry Handler empty for now.")

    def get_smallest_dx(self):
        """
        Returns (in code units) the smallest cell size in the simulation.
        """
        return (
            self.dataset.domain_width
            / (self.dataset.domain_dimensions * 2 ** (self.max_level))
        ).min()

    def _get_particle_type_counts(self):
        # this could be done in the artio C interface without creating temporary
        # arrays but I don't want to touch that code
        # if a future brave soul wants to try, take a look at
        # `read_sfc_particles` in _artio_caller.pyx
        result = {}
        ad = self.ds.all_data()
        for ptype in self.ds.particle_types_raw:
            result[ptype] = ad[ptype, "PID"].size
        return result

    def convert(self, unit):
        return self.dataset.conversion_factors[unit]

    def find_max(self, field, finest_levels=3):
        """
        Returns (value, center) of location of maximum for a given field.
        """
        if (field, finest_levels) in self._max_locations:
            return self._max_locations[(field, finest_levels)]
        mv, pos = self.find_max_cell_location(field, finest_levels)
        self._max_locations[(field, finest_levels)] = (mv, pos)
        return mv, pos

    def find_max_cell_location(self, field, finest_levels=3):
        source = self.all_data()
        if finest_levels is not False:
            source.min_level = self.max_level - finest_levels
        mylog.debug("Searching for maximum value of %s", field)
        max_val, mx, my, mz = source.quantities["MaxLocation"](field)
        mylog.info("Max Value is %0.5e at %0.16f %0.16f %0.16f", max_val, mx, my, mz)
        self.ds.parameters[f"Max{field}Value"] = max_val
        self.ds.parameters[f"Max{field}Pos"] = f"{mx, my, mz}"
        return max_val, np.array((mx, my, mz), dtype="float64")

    def _detect_output_fields(self):
        self.fluid_field_list = self._detect_fluid_fields()
        self.particle_field_list = self._detect_particle_fields()
        self.field_list = self.fluid_field_list + self.particle_field_list
        mylog.debug("Detected fields: %s", (self.field_list,))

    def _detect_fluid_fields(self):
        return [("artio", f) for f in self.ds.artio_parameters["grid_variable_labels"]]

    def _detect_particle_fields(self):
        fields = set()
        for i, ptype in enumerate(self.ds.particle_types):
            if ptype == "all":
                break  # This will always be after all intrinsic
            for fname in self.ds.particle_variables[i]:
                fields.add((ptype, fname))
        return list(fields)

    def _identify_base_chunk(self, dobj):
        if getattr(dobj, "_chunk_info", None) is None:
            try:
                all_data = all(dobj.left_edge == self.ds.domain_left_edge) and all(
                    dobj.right_edge == self.ds.domain_right_edge
                )
            except Exception:
                all_data = False
            base_region = getattr(dobj, "base_region", dobj)
            sfc_start = getattr(dobj, "sfc_start", None)
            sfc_end = getattr(dobj, "sfc_end", None)
            nz = getattr(dobj, "_num_zones", 0)
            if all_data:
                mylog.debug("Selecting entire artio domain")
                list_sfc_ranges = self.ds._handle.root_sfc_ranges_all(
                    max_range_size=self.max_range
                )
            elif sfc_start is not None and sfc_end is not None:
                mylog.debug("Restricting to %s .. %s", sfc_start, sfc_end)
                list_sfc_ranges = [(sfc_start, sfc_end)]
            else:
                mylog.debug("Running selector on artio base grid")
                list_sfc_ranges = self.ds._handle.root_sfc_ranges(
                    dobj.selector, max_range_size=self.max_range
                )
            ci = []
            # v = np.array(list_sfc_ranges)
            # list_sfc_ranges = [ (v.min(), v.max()) ]
            for start, end in list_sfc_ranges:
                if (start, end) in self.range_handlers.keys():
                    range_handler = self.range_handlers[(start, end)]
                else:
                    range_handler = ARTIOSFCRangeHandler(
                        self.ds.domain_dimensions,
                        self.ds.domain_left_edge,
                        self.ds.domain_right_edge,
                        self.ds._handle,
                        start,
                        end,
                    )
                    range_handler.construct_mesh()
                    self.range_handlers[(start, end)] = range_handler
                if nz != 2:
                    ci.append(
                        ARTIORootMeshSubset(
                            base_region,
                            start,
                            end,
                            range_handler.root_mesh_handler,
                            self.ds,
                        )
                    )
                if nz != 1 and range_handler.total_octs > 0:
                    ci.append(
                        ARTIOOctreeSubset(
                            base_region,
                            start,
                            end,
                            range_handler.octree_handler,
                            self.ds,
                        )
                    )
            dobj._chunk_info = ci
            if len(list_sfc_ranges) > 1:
                mylog.info("Created %d chunks for ARTIO", len(list_sfc_ranges))
        dobj._current_chunk = list(self._chunk_all(dobj))[0]

    def _data_size(self, dobj, dobjs):
        size = 0
        for d in dobjs:
            size += d.data_size
        return size

    def _chunk_all(self, dobj):
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        yield YTDataChunk(dobj, "all", oobjs, None, cache=True)

    def _chunk_spatial(self, dobj, ngz, preload_fields=None):
        if ngz > 0:
            raise NotImplementedError
        sobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for og in sobjs:
            if ngz > 0:
                g = og.retrieve_ghost_zones(ngz, [], smoothed=True)
            else:
                g = og
            yield YTDataChunk(dobj, "spatial", [g], None, cache=True)

    def _chunk_io(self, dobj, cache=True, local_only=False):
        # _current_chunk is made from identify_base_chunk
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for chunk in oobjs:
            yield YTDataChunk(dobj, "io", [chunk], None, cache=cache)

    def _read_fluid_fields(self, fields, dobj, chunk=None):
        if len(fields) == 0:
            return {}, []
        if chunk is None:
            self._identify_base_chunk(dobj)
        fields_to_return = {}
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        fields_to_return = self.io._read_fluid_selection(
            self._chunk_io(dobj), dobj.selector, fields_to_read
        )
        return fields_to_return, fields_to_generate

    def _icoords_to_fcoords(
        self,
        icoords: np.ndarray,
        ires: np.ndarray,
        axes: Optional[Tuple[int, ...]] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Accepts icoords and ires and returns appropriate fcoords and fwidth.
        Mostly useful for cases where we have irregularly spaced or structured
        grids.
        """
        dds = self.ds.domain_width[(axes,)] / (
            self.ds.domain_dimensions[axes,] * self.ds.refine_by ** ires[:, None]
        )
        pos = (0.5 + icoords) * dds + self.ds.domain_left_edge[axes,]
        return pos, dds


class ARTIODataset(Dataset):
    _handle = None
    _index_class = ARTIOIndex
    _field_info_class = ARTIOFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="artio",
        storage_filename=None,
        max_range=1024,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        if self._handle is not None:
            return
        self.max_range = max_range
        self.fluid_types += ("artio",)
        self._filename = filename
        self._fileset_prefix = filename[:-4]
        self._handle = artio_fileset(bytes(self._fileset_prefix, "utf-8"))
        self.artio_parameters = self._handle.parameters
        # Here we want to initiate a traceback, if the reader is not built.
        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        self.storage_filename = storage_filename

    def _set_code_unit_attributes(self):
        setdefaultattr(self, "mass_unit", self.quan(self.parameters["unit_m"], "g"))
        setdefaultattr(self, "length_unit", self.quan(self.parameters["unit_l"], "cm"))
        setdefaultattr(self, "time_unit", self.quan(self.parameters["unit_t"], "s"))
        setdefaultattr(self, "velocity_unit", self.length_unit / self.time_unit)

    def _parse_parameter_file(self):
        # hard-coded -- not provided by headers
        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = "artio"
        self.parameters["Time"] = 1.0  # default unit is 1...

        # read header

        self.num_grid = self._handle.num_grid
        self.domain_dimensions = np.ones(3, dtype="int32") * self.num_grid
        self.domain_left_edge = np.zeros(3, dtype="float64")
        self.domain_right_edge = np.ones(3, dtype="float64") * self.num_grid

        # TODO: detect if grid exists
        self.min_level = 0  # ART has min_level=0
        self.max_level = self.artio_parameters["grid_max_level"][0]

        # TODO: detect if particles exist
        if self._handle.has_particles:
            self.num_species = self.artio_parameters["num_particle_species"][0]
            self.particle_variables = [
                ["PID", "SPECIES"] for i in range(self.num_species)
            ]

            # If multiple N-BODY species exist, they all have the same name,
            # which can lead to conflict if not renamed
            # A particle union will be created later to hold all N-BODY
            # particles and will take the name "N-BODY"
            labels = self.artio_parameters["particle_species_labels"]
            if labels.count("N-BODY") > 1:
                for species, label in enumerate(labels):
                    if label == "N-BODY":
                        labels[species] = f"N-BODY_{species}"

            self.particle_types_raw = self.artio_parameters["particle_species_labels"]
            self.particle_types = tuple(self.particle_types_raw)

            for species in range(self.num_species):
                # Mass would be best as a derived field,
                # but wouldn't detect under 'all'
                label = self.artio_parameters["particle_species_labels"][species]
                if "N-BODY" in label:
                    self.particle_variables[species].append("MASS")

                if self.artio_parameters["num_primary_variables"][species] > 0:
                    self.particle_variables[species].extend(
                        self.artio_parameters[
                            "species_%02d_primary_variable_labels" % (species,)
                        ]
                    )
                if self.artio_parameters["num_secondary_variables"][species] > 0:
                    self.particle_variables[species].extend(
                        self.artio_parameters[
                            "species_%02d_secondary_variable_labels" % (species,)
                        ]
                    )

        else:
            self.num_species = 0
            self.particle_variables = []
            self.particle_types = ()
        self.particle_types_raw = self.particle_types

        self.current_time = self.quan(
            self._handle.tphys_from_tcode(self.artio_parameters["tl"][0]), "yr"
        )

        # detect cosmology
        if "abox" in self.artio_parameters:
            self.cosmological_simulation = True

            abox = self.artio_parameters["abox"][0]
            self.omega_lambda = self.artio_parameters["OmegaL"][0]
            self.omega_matter = self.artio_parameters["OmegaM"][0]
            self.hubble_constant = self.artio_parameters["hubble"][0]
            self.current_redshift = 1.0 / self.artio_parameters["auni"][0] - 1.0
            self.current_redshift_box = 1.0 / abox - 1.0

            self.parameters["initial_redshift"] = (
                1.0 / self.artio_parameters["auni_init"][0] - 1.0
            )
            self.parameters["CosmologyInitialRedshift"] = self.parameters[
                "initial_redshift"
            ]

            self.parameters["unit_m"] = self.artio_parameters["mass_unit"][0]
            self.parameters["unit_t"] = (
                self.artio_parameters["time_unit"][0] * abox**2
            )
            self.parameters["unit_l"] = self.artio_parameters["length_unit"][0] * abox

            if self.artio_parameters["DeltaDC"][0] != 0:
                mylog.warning(
                    "DeltaDC != 0, which implies auni != abox. "
                    "Be sure you understand which expansion parameter "
                    "is appropriate for your use! (Gnedin, Kravtsov, & Rudd 2011)"
                )
        else:
            self.cosmological_simulation = False

            self.parameters["unit_l"] = self.artio_parameters["length_unit"][0]
            self.parameters["unit_t"] = self.artio_parameters["time_unit"][0]
            self.parameters["unit_m"] = self.artio_parameters["mass_unit"][0]

        # hard coded assumption of 3D periodicity
        self._periodicity = (True, True, True)

    def create_field_info(self):
        super().create_field_info()
        # only make the particle union if there are multiple DM species.
        # If there are multiple, "N-BODY_0" will be the first species. If there
        # are not multiple, they will be all under "N-BODY"
        if "N-BODY_0" in self.particle_types_raw:
            dm_labels = [
                label for label in self.particle_types_raw if "N-BODY" in label
            ]
            # Use the N-BODY label for the union to be consistent with the
            # previous single mass N-BODY case, where this label was used for
            # all N-BODY particles by default
            pu = ParticleUnion("N-BODY", dm_labels)
            self.add_particle_union(pu)

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        # a valid artio header file starts with a prefix and ends with .art
        name, _, ext = filename.rpartition(".")
        if ext != "art":
            return False
        return artio_is_valid(bytes(name, "utf-8"))



import sys

import pytest

import yt
from yt.data_objects.static_output import Dataset
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.object_registries import output_type_registry


class MockEntryPoint:
    @classmethod
    def load(cls):
        class MockHierarchy(GridIndex):
            grid = None

        class ExtDataset(Dataset):
            _index_class = MockHierarchy

            def _parse_parameter_file(self):
                self.current_time = 1.0
                self.cosmological_simulation = 0

            def _set_code_unit_attributes(self):
                self.length_unit = self.quan(1.0, "code_length")
                self.mass_unit = self.quan(1.0, "code_mass")
                self.time_unit = self.quan(1.0, "code_time")

            @classmethod
            def _is_valid(cls, filename, *args, **kwargs):
                return filename.endswith("mock")


@pytest.fixture()
def mock_external_frontend(monkeypatch):
    def mock_entry_points(group=None):
        if sys.version_info >= (3, 10):
            return [MockEntryPoint]
        else:
            return {"yt.frontends": [MockEntryPoint]}

    monkeypatch.setattr(yt.loaders, "entry_points", mock_entry_points)
    assert "ExtDataset" not in output_type_registry

    yield

    assert "ExtDataset" in output_type_registry
    # teardown to avoid test pollution
    output_type_registry.pop("ExtDataset")


@pytest.mark.usefixtures("mock_external_frontend")
def test_external_frontend(tmp_path):
    test_file = tmp_path / "tmp.mock"
    test_file.write_text("")  # create the file
    assert test_file.is_file()

    ds = yt.load(test_file)
    assert "ExtDataset" in ds.__class__.__name__
